#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Staging
=======

Manage object staging for pushing to a remote registry from the local registry

Contents
========

Classes
-------

    Staging - class handles staging of objects ready to be synchronised

"""

__date__ = "2021-07-13"

import os
import typing
import yaml
import logging

import fair.common as fdp_com
import fair.run as fdp_run
import fair.registry.requests as fdp_req
import fair.exceptions as fdp_exc
import fair.configuration as fdp_conf


class Stager:
    """
    The stager handles staging of jobs and other objects for synchronising
    between local and remote registries. In a manner similar to git, objects
    such as jobs can be tracked but not staged, for example if a job is a trial
    or has issues.
    """
    _logger = logging.getLogger('FAIRDataPipeline.Staging')
    def __init__(self, repo_root: str) -> None:
        """Create a new stager instance for a given FAIR repository

        Parameters
        ----------
        repo_root : str, optional
            the root of the FAIR repository, by default find the root from the
            current working directory
        """
        self._root = repo_root
        self._staging_file = fdp_com.staging_cache(self._root)
        self._logger.debug("Creating stager for FAIR repository '%s'", repo_root)

    def initialise(self) -> None:
        """Initialise the stager, creating a staging cache file if one does not exist"""
        # Only create the staging file if one is not already present within the
        # specified directory
        if not os.path.exists(self._staging_file):
            self._logger.debug("Creating new staging cache file")
            # If the stager is called before the rest of the directory tree
            # has been created make the parent directories first
            if not os.path.exists(os.path.dirname(self._staging_file)):
                os.makedirs(os.path.dirname(self._staging_file), exist_ok=True)
            self._create_staging_file()
        else:
            self._logger.debug("Existing staging cache found")


    def _create_file_label(self, file_to_stage: str) -> str:
        return os.path.relpath(
            file_to_stage,
            os.path.dirname(self._staging_file),
        )

    def _create_staging_file(self) -> None:
        _staging_dict = {
            'job': {},
            'file': {}
        }
        yaml.dump(_staging_dict, open(self._staging_file, 'w'))

    def change_job_stage_status(self, job_id: str, stage: bool = True) -> None:
        """Stage a local code job ready to be pushed to the remote registry

        Parameters
        ----------
        job_id : str
            a valid uuid for the given job
        stage : bool, optional
            whether job is staged, default True
        """
        self._logger.debug("Setting job '%s' status to staged=%s", job_id, stage)

        if not os.path.exists(self._staging_file):
            raise fdp_exc.FileNotFoundError(
                "Failed to update tracking, expected staging file"
                f" '{self._staging_file}' but it does not exist"
            )

        # Open the staging dictionary first
        _staging_dict = yaml.safe_load(open(self._staging_file))

        # When a job is completed by a language implementation the CLI should
        # have already registered it into staging with a status of staged=False
        if not fdp_run.get_job_dir(job_id):
            raise fdp_exc.StagingError(
                f"Failed to recognise job with ID '{job_id}'"
            )

        _staging_dict['job'][job_id] = stage

        with open(self._staging_file, 'w') as f:
            yaml.dump(_staging_dict, f)


    def find_registry_entry_for_file(self, local_uri: str, file_path: str) -> str:
        """Performs a rough search for a file in the local registry

        Parameters
        ----------
        file_path : str
            local file system path

        Returns
        -------
        str
            URL of local registry entry
        """
        self._logger.debug(
            "Retrieving registry entry for file '%s'",
            file_path
        )

        # Will search storage locations for a similar path by using
        # parent_directory/file_name
        _obj_type = 'storage_location'

        _results = fdp_req.get(local_uri, _obj_type, params={"path": file_path})

        if not _results:
            raise fdp_exc.StagingError(
                "Failed to find local registry entry for file "
                f"'{file_path}'"
            )

        if len(_results) > 1:
            raise fdp_exc.StagingError(
                "Expected single result for local registry entry relating to "
                f" file '{file_path} but got {len(_results)}"
            )
        
        self._logger.debug("Found config.yaml URL: %s", _results[0])

        return _results[0]

    
    def _get_code_run_entries(
        self,
        local_uri: str,
        job_dir: str
        ) -> typing.List[str]:
        """Retrieve code_run URL list from a given CLI run directory

        Parameters
        ----------
        local_uri : str
            local registry endpoint
        job_dir : str
            CLI run directory

        Returns
        -------
        typing.List[str]

        Raises
        ------
        fdp_exc.ImplementationError
            If the expected registry entries have not been created by an API
            implementation
        """
        self._logger.debug(
            "Retrieving code run URL for job '%s'",
            os.path.basename(job_dir)
        )
        # Check if any code_runs are present for the given job
        _code_run_file = os.path.join(job_dir, 'coderuns.txt')
        _code_run_urls = []

        if ( os.path.exists(_code_run_file) and
            open(_code_run_file).read().strip()
        ):
            self._logger.debug("Found coderuns file, extracting runs")
            _runs = [i.strip() for i in open(_code_run_file).readlines()]
            
            for run in _runs:
                _results = fdp_req.get(
                    local_uri,
                    'code_run',
                    params={"uuid": run}
                )

                if not _results:
                    raise fdp_exc.ImplementationError(
                        "Expected code_run with uuid "
                        f"'{run}' in local registry, but no result found."
                    )
                
                _code_run_urls.append(_results[0]["url"])
            
        return _code_run_urls

    def _get_written_obj_entries(
        self,
        local_uri: str,
        config_dict: typing.Dict
    ):
        if 'write' not in config_dict:
            return []

        _data_product_urls: typing.List[str] = []
        
        for write_obj in config_dict['write']:
            _data_product = write_obj['data_product']
            _results = fdp_req.get(
                local_uri,
                'data_product',
                params={"name": _data_product}
            )

            if not _results:
                raise fdp_exc.InternalError(
                    "Expected data_product "
                    f"'{_data_product}' in local registry, but no result found."
                )
            
            _data_product_urls.append(_results[0]["url"])
        
        return _data_product_urls
    
    def get_job_data(
            self, local_uri, identifier: str
        ) -> typing.Dict[str, str]:
        self._logger.debug("Fetching job data for job %s", identifier)
        # Firstly find the job directory
        _directory = fdp_run.get_job_dir(identifier)
        if not _directory:
            raise fdp_exc.StagingError(
                f"Could not retrieve directory for job '{identifier}'"
            )
        
        # Check for a config.yaml file
        _config_yaml = os.path.join(_directory, fdp_com.USER_CONFIG_FILE)
        if not os.path.exists(_config_yaml):
            raise fdp_exc.FileNotFoundError(
                f"Cannot stage job '{identifier}' "
                f"Expected config.yaml in '{_directory}'"
            )

        # Find this config.yaml on the local registry, this involves
        # firstly getting the path commencing from the 'jobs' folder
        _config_rel_path = _config_yaml.split(fdp_com.JOBS_DIR)[1]
        _config_rel_path = f'{fdp_com.JOBS_DIR}{_config_rel_path}'

        _config_url = self.find_registry_entry_for_file(
            fdp_conf.get_local_uri(),
            _config_rel_path
        )["url"]

        # Check for job script file
        _config_yaml = os.path.join(_directory, fdp_com.USER_CONFIG_FILE)
        if not os.path.exists(_config_yaml):
            raise fdp_exc.FileNotFoundError(
                f"Cannot stage job '{identifier}' "
                f"Expected config.yaml in '{_directory}'"
            )

        # Find this job script on the local registry, as the script
        # can have any name obtain this information from the config.yaml
        _config_dict = yaml.safe_load(open(_config_yaml))

        if (
            'run_metadata' not in _config_dict
            or 'script_path' not in _config_dict['run_metadata']
        ):
            _script_url  = None 
        else:      

            # Find the relevant script path on the local registry, this involves
            # firstly getting the path commencing from the 'jobs' folder
            self._logger.debug("Finding job script within local registry")
            _script_path = _config_dict['run_metadata']['script_path']
            _rel_script_path = _script_path.split(fdp_com.JOBS_DIR)[1]
            _rel_script_path = f'{fdp_com.JOBS_DIR}{_rel_script_path}'

            _script_url = self.find_registry_entry_for_file(
                local_uri,
                _rel_script_path
            )["url"]
            self._logger.debug("Script URL: %s", _script_url)

        self._logger.debug("Retrieving code runs and written objects")

        _code_run_urls = self._get_code_run_entries(local_uri, _directory)
        _user_written_obj_urls = self._get_written_obj_entries(
            local_uri, _config_dict
        )

        return {
            "jobs": _code_run_urls,
            "user_written_objects": _user_written_obj_urls,
            "config_file": _config_url,
            "script_file": _script_url
        }

    
    def remove_staging_entry(
        self, identifier: str, stage_type: str = "job"
        ) -> None:
        """Remove an item of type 'stage_type' from staging

        Parameters
        ----------
            identifier : str
                name or ID of item
            stage_type: str, optional
                type of stage item either job (default) or file
        """
        # Open the staging dictionary first
        _staging_dict = yaml.safe_load(open(self._staging_file))

        if stage_type not in _staging_dict:
            raise fdp_exc.StagingError(
                f"Cannot remove staging item of unrecognised type '{stage_type}'"
            )

        if identifier not in _staging_dict[stage_type]:
            raise fdp_exc.StagingError(
                f"Cannot remove item '{identifier}' of stage type '{stage_type}', "
                "item is not in staging."
            )
        
        del _staging_dict[stage_type][identifier]

        with open(self._staging_file, 'w') as f:
            yaml.dump(_staging_dict, f)

    def get_item_list(
        self, staged: bool = True, stage_type: str = "job"
        ) -> typing.List[str]:
        """Returns a list of items of type 'stage_type' which are staged/unstaged
        
        Parameters
        ----------
            staged : bool
                list staged/unstaged items
            stage_type : str, optional
                type of stage item either job (default) or file
        """
        _staging_dict = yaml.safe_load(open(self._staging_file))

        if stage_type not in _staging_dict:
            raise fdp_exc.StagingError(
                f"Cannot remove staging item of unrecognised type '{stage_type}'"
            )

        return [k for k, v in _staging_dict[stage_type].items() if v == staged]



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Run History
===========

Methods relating to the summary of jobs commenced within the given FAIR-CLI
repository. Allowing user to view any stdout from these.

Contents
========

Functions
---------

    history_directory - returns the current repository logs directory
"""
import os
import glob
import click
import rich

import fair.common as fdp_com
import fair.run as fdp_run
import fair.exceptions as fdp_exc
import fair.templates as fdp_tpl


def history_directory(repo_loc: str) -> str:
    """Retrieve the directory containing job logs for the specified repository

    Parameters
    ----------
    repo_loc : str
        FAIR-CLI repository path

    Returns
    -------
    str
        location of the job logs directory
    """
    return os.path.join(
        fdp_com.find_fair_root(repo_loc), fdp_com.FAIR_FOLDER, "logs"
    )


def show_job_log(repo_loc: str, job_id: str) -> str:
    """Show the log from a given job

    Parameters
    ----------
    repo_loc : str
        FAIR-CLI repository path
    job_id : str
        SHA identifier for the job

    Returns
    -------
    str
        log file location for the given job
    """
    _sorted_time_dirs = sorted(
        glob.glob(os.path.join(fdp_com.default_jobs_dir(), "*")),
        reverse=True
    )

    _log_files = [
        os.path.join(
            history_directory(repo_loc),
            f'job_{os.path.basename(i)}.log'
        )
        for i in _sorted_time_dirs
    ]

    for job_dir, _log_file in zip(_sorted_time_dirs, _log_files):
        # Use the timestamp directory name for the hash
        _job_id = fdp_run.get_job_hash(job_dir)

        if _job_id[: len(job_id)] == job_id:
            with open(_log_file) as f:
                click.echo(f.read())
            _jobs_list = os.path.join(job_dir, 'coderuns.txt')

            # Check if a code runs file exists for the given job and also
            # print the list of code_run uuids created in the registry
            if os.path.exists(_jobs_list):
                click.echo("Related Code Runs: ")
                click.echo('\n\t- '.join(open(_jobs_list).readlines()))

            return _log_file

    raise fdp_exc.FileNotFoundError(
        f"Could not find job matching id '{job_id}'"
    )


def show_history(repo_loc: str, length: int = 10) -> None:
    """Show job history by time sorting log outputs, display metadata

    Parameters
    ----------
    repo_loc : str
        FAIR-CLI repository path
    length : int, optional
        max number of entries to display, by default 10
    """

    _sorted_time_dirs = sorted(
        glob.glob(os.path.join(fdp_com.default_jobs_dir(), "*")),
        reverse=True
    )

    _log_files = [
        os.path.join(
            history_directory(repo_loc),
            f'job_{os.path.basename(i)}.log'
        )
        for i in _sorted_time_dirs
    ]

    # Iterate through the logs printing out the job author
    for i, (job_dir, _log_file) in enumerate(
        zip(_sorted_time_dirs, _log_files)
    ):
        _job_id = fdp_run.get_job_hash(job_dir)
        if not os.path.exists(_log_file):
            raise fdp_exc.FileNotFoundError(
                f"Cannot open log for job '{_job_id}'"
            )
        with open(_log_file) as f:
            _log_lines = f.readlines()
            _metadata = []
            for line in _log_lines:
                if '------- time taken ' in line:
                    break
                if ' = ' in line:
                    _metadata.append(line)
        if not _metadata:
            continue
        _metadata = [i for i in _metadata if i.strip()]
        _user_lines = [i for i in _metadata if 'Author' in i]
        if not _user_lines:
            raise fdp_exc.InternalError(
                "Failed to retrieve author information from log "
                f"for job '{_job_id}'"
            )
        _date_lines = [i for i in _metadata if 'Commenced' in i]
        if not _date_lines:
            raise fdp_exc.InternalError(
                "Failed to retrieve date information from log "
                f"for job '{_job_id}'"
            )
        _date = _date_lines[0].split('=')[1].strip()
        _user = _user_lines[0].split('=')[1].strip()
        _name = _user.split("<")[0].strip()
        _email = _user.replace(_name, "").strip()
        _meta = {
            "sha": _job_id,
            "user": _name,
            "user_email": _email,
            "datetime": _date,
        }
        rich.print(fdp_tpl.hist_template.render(**_meta))

        if i == length:
            return



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Register Key Methods
====================

Substitute/fetch data relating to 'register' entries in the
`config.yaml` user config


Contents
========

Functions
-------

"""

__date__ = "2021-08-16"

import copy
import os
import typing
import urllib.parse
import shutil
import logging
import requests

import fair.exceptions as fdp_exc
import fair.registry.requests as fdp_req
import fair.registry.storage as fdp_store
import fair.registry.versioning as fdp_ver
import fair.configuration as fdp_conf


def fetch_registrations(
    local_uri: str,
    repo_dir: str,
    write_data_store: str,
    user_config_register: typing.Dict,
    ) -> typing.Dict:
    """As part of running 'pull' fetch all items listed under 'register'

    Parameters
    ----------
    repo_dir : str
        FAIR project repository
    cfg : typing.Dict
        Dict containing working config       

    Returns
    -------
    typing.List[str]
        list of registered object URLs
    """
    _expected_keys = [
        "root",
        "path",
        "file_type",
        "primary",
        "version",
        "public"
    ]
    _logger = logging.getLogger("FAIRDataPipeline.Run")

    _stored_objects: typing.List[str] = []

    for entry in user_config_register:
        for key in _expected_keys:
            if key not in entry and key not in entry['use']:
                raise fdp_exc.UserConfigError(
                    f"Expected key '{key}' in 'register' item"
                )

        _identifier: str = entry['identifier'] if 'identifier' in entry else ''
        _unique_name: str = entry['unique_name'] if 'unique_name' in entry else ''

        _data_product = None
        _external_object = None

        _search_data = {}

        if 'data_product' in entry:
            _data_product: str = entry['use']['data_product']
        elif 'external_object' in entry:
            _external_object: str = entry['use']['data_product']

        if not _external_object and not _data_product:
            raise fdp_exc.UserConfigError(
                "Expected either 'data_product' or 'external_object' in 'register' item"
            )

        elif _external_object and _data_product:
            raise fdp_exc.UserConfigError(
                "Only one type may be provided (data_product/external_object)"
            )
        elif _external_object:
            _name = entry['use']['data_product']
            _obj_type = 'external_object'
            #TODO: This doesn't work because of a mismatch with spaces in alternate_identifier, perhaps?
            if 'unique_name' in entry and 'alternate_identifier_type' in entry:
#                _search_data['alternate_identifier'] = entry['unique_name']
                _search_data['alternate_identifier_type'] = entry['alternate_identifier_type']
            elif 'identifier' in entry:
                _search_data['identifier'] = entry['identifier']
            else:
                raise fdp_exc.UserConfigError(
                    "Expected either 'identifier', or 'unique_name' and "
                    f"'alternate_identifier_type' in external object '{_name}'"
                )
            _search_data['namespace'] = entry['use']['namespace']
            _search_data['data_product'] = entry['use']['data_product']
        else:
            _name = entry['use']['data_product']
            _obj_type = 'data_product'
            _search_data = {"name": _name}

        _search_data['version'] = entry['use']['version']
        _namespace = entry['use']['namespace']

        if not _identifier and not _unique_name:
            raise fdp_exc.UserConfigError(
                "Expected either 'unique_name' or 'identifier' in 'register' item"
            )

        elif _identifier and _unique_name:
            raise fdp_exc.UserConfigError(
                "Only one unique identifier may be provided (doi/unique_name)"
            )

        if 'cache' in entry: # Do we have a local cache already?
            _temp_data_file = entry['cache']
        else: # Need to download it
            _root, _path = entry["root"], entry['path']

            # Encode the path first
            _path = urllib.parse.quote_plus(_path)
            _url = f"{_root}{_path}"
            try:
                _temp_data_file = fdp_req.download_file(_url)
            except requests.HTTPError as r_in:
                raise fdp_exc.UserConfigError(
                    f"Failed to fetch item '{_url}' with exit code "
                    f"{r_in.response}"
                )

        # Need to fix the path for Windows
        if os.path.sep != '/':
            _name = _name.replace('/', os.path.sep)

        _local_dir = os.path.join(
            write_data_store, _namespace, _name
        )

        # Check if the object is already present on the local registry
        _is_present = fdp_store.check_if_object_exists(
            local_uri=local_uri,
            file_loc=_temp_data_file,
            obj_type=_obj_type,
            search_data=_search_data
        )

        # Hash matched version already present
        if _is_present == "hash_match":
            _logger.debug(
                f"Skipping item '{_name}' as a hash matched entry is already"
                " present with this name"
            )
            os.remove(_temp_data_file)
            continue

        # Item found but not hash matched retrieve a version number
        elif _is_present != "absent":
            _results = _is_present
            _user_version = fdp_ver.get_correct_version(
                results_list=_results,
                free_write=True,
                version=entry['use']['version']
            )
            _logger.debug("Found results for %s", str(_results))
        else:
            _user_version = fdp_ver.get_correct_version(
                results_list=None,
                free_write=True,
                version=entry['use']['version']
            )
            _logger.debug("Found nothing for %s", str(_search_data))

        # Create object location directory, ignoring if already present
        # as multiple version files can exist
        os.makedirs(_local_dir, exist_ok=True)

        _local_file = os.path.join(
            _local_dir,
            f"{_user_version}.{entry['file_type']}"
        )

        # Copy the temporary file into the data store
        # then remove temporary file to save space
        shutil.copy(_temp_data_file, _local_file)
        if 'cache' not in entry:
            os.remove(_temp_data_file)

        if 'public' in entry:
            _public = entry['public']

        data=copy.deepcopy(entry)

        data['namespace_name'] = entry['use']['namespace']

        _file_url = fdp_store.store_data_file(
            uri=local_uri,
            repo_dir=repo_dir,
            data=data,
            local_file=_local_file,
            write_data_store=write_data_store,
            public=_public
        )

        _stored_objects.append(_file_url)
    
    return _stored_objects



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Registry Server Control
=======================

Contains methods for controlling the starting and stopping of the local
FAIR Data Pipeline registry via the CLI.

Contents
========

Functions
---------

    check_server_running - confirm that the server is running locally

"""

__date__ = "2021-06-28"

import os
import subprocess
import glob
import shutil
import sys
import pathlib
import enum
import logging
import venv
import typing
from gitdb.db import ref
import requests
import platform
import git
import yaml

import fair.common as fdp_com
import fair.exceptions as fdp_exc
import fair.configuration as fdp_conf
import fair.registry.storage as fdp_store
import fair.registry.requests as fdp_req
import fair.utilities as fdp_util

FAIR_REGISTRY_REPO = "https://github.com/FAIRDataPipeline/data-registry.git"

DEFAULT_REGISTRY_DOMAIN = "https://data.scrc.uk/"
REGISTRY_INSTALL_URL = "https://data.scrc.uk/static/localregistry.sh"
DEFAULT_REGISTRY_LOCATION = os.path.join(pathlib.Path().home(), fdp_com.FAIR_FOLDER, 'registry')
DEFAULT_LOCAL_REGISTRY_URL = "http://127.0.0.1:8000/api/"


def django_environ(environ: typing.Dict = os.environ):
    _environ = environ.copy()
    _environ['DJANGO_SETTINGS_MODULE'] = 'drams.local-settings'
    _environ['DJANGO_SUPERUSER_USERNAME'] = 'admin'
    _environ['DJANGO_SUPERUSER_PASSWORD'] = 'admin'
    return _environ


class SwitchMode(enum.Enum):
    """Server access mode

    The server can be launched either by the user or by the CLI itself. In the
    latter case it will be shut down after use if no other instances are using it.
    However if the user manually starts it themselves then the CLI will not
    alter the state. This class handles all stop/start request possibilities.
    """

    USER_START = 0
    USER_STOP = 1
    CLI = 2
    FORCE_STOP = 3
    NO_SERVER = 4


def check_server_running(local_uri: str = None) -> bool:
    """Check the state of server

    Parameters
    ----------
    local_uri : str
        local registyr endpoint

    Returns
    -------
    bool
        whether server is running
    """
    logger = logging.getLogger('FAIRDataPipeline.Server')

    if not local_uri:
        local_uri = fdp_conf.get_local_uri()

    logger.debug("Checking if server is running on '%s'", local_uri)

    try:
        _status_code = requests.get(local_uri).status_code
        assert _status_code == 200
        return True
    except (requests.exceptions.ConnectionError, AssertionError):
        return False


def launch_server(local_uri: str = None, registry_dir: str = None, verbose: bool = False) -> int:
    """Start the registry server.

    Parameters
    ----------
    verbose : bool, optional
        show registry start output, by default False
    """
    logger = logging.getLogger('FAIRDataPipeline.Server')

    if not registry_dir:
        registry_dir = fdp_com.registry_home()

    _server_start_script = os.path.join(
        registry_dir, "scripts", "start_fair_registry"
    )

    if platform.system() == "Windows":
        _server_start_script += '_windows.bat'

    if not os.path.exists(_server_start_script):
        raise fdp_exc.RegistryError(
            f"Failed to find local registry executable '{_server_start_script}',"
            " is the FAIR data pipeline properly installed on this system?"
        )

    _cmd = [_server_start_script, '-p', f'{fdp_conf.get_local_port(local_uri)}']

    logger.debug("Launching server with command '%s'", ' '.join(_cmd))

    _start = subprocess.Popen(
        _cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=False,
    )

    if verbose and platform.system() != "Windows":
        for c in iter(lambda: _start.stdout.read(1), b""):
            sys.stdout.buffer.write(c)

    _start.wait()

    if not check_server_running(local_uri):
        raise fdp_exc.RegistryError(
            "Failed to start local registry, no response from server"
        )


def stop_server(
    registry_dir: str = None, local_uri: str = None,
    force: bool = False, verbose: bool = False
) -> None:
    """Stops the FAIR Data Pipeline local server

    Parameters
    ----------
    force : bool, optional
        whether to force server shutdown if it is being used
    """
    logger = logging.getLogger('FAIRDataPipeline.Server')

    registry_dir = registry_dir or fdp_com.registry_home()
    _session_port_file = os.path.join(registry_dir, 'session_port.log')

    if not os.path.exists(_session_port_file):
        raise fdp_exc.FileNotFoundError(
            "Failed to retrieve current session port from file "
            f"'{_session_port_file}', file does not exist"
        )
    
    # If there are no session cache files shut down server
    _run_files = glob.glob(os.path.join(fdp_com.session_cache_dir(), "*.run"))
    if not force and _run_files:
        raise fdp_exc.RegistryError(
            "Could not stop registry server, processes still running."
        )

    _server_stop_script = os.path.join(
        registry_dir, "scripts", "stop_fair_registry"
    )

    if platform.system() == "Windows":
        _server_stop_script += '_windows.bat'

    if not os.path.exists(_server_stop_script):
        raise fdp_exc.RegistryError(
            f"Failed to find local registry executable '{_server_stop_script}',"
            " is the FAIR data pipeline properly installed on this system?"
        )

    logger.debug("Stopping local registry server.")

    _stop = subprocess.Popen(
        _server_stop_script,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=False,
    )

    if verbose:
        for c in iter(lambda: _stop.stdout.read(1), b""):
            sys.stdout.buffer.write(c)

    _stop.wait()

    if check_server_running(local_uri):
        raise fdp_exc.RegistryError("Failed to stop registry server.")


def rebuild_local(python: str, install_dir: str = None, silent: bool = False):
    """Rebuild the local Django registry

    Parameters
    ----------
    python : str
        python binary of virtual environment
    install_dir : str, optional
        location of registry installation
    silent : bool, optional
        run in silent mode, by default False
    """

    logger = logging.getLogger('FAIRDataPipeline.Server')
    logger.debug("Rebuilding local registry")
    if not install_dir:
        install_dir = DEFAULT_REGISTRY_LOCATION
    
    _migration_files = glob.glob(
        os.path.join(install_dir, '*', 'migrations', '*.py*')
    )

    logger.debug("Removing migration files: %s", _migration_files)

    for mf in _migration_files:
        os.remove(mf)

    _db_file = os.path.join(install_dir, 'db.sqlite3')

    logger.debug("Removing SQL database file")

    if os.path.exists(_db_file):
        os.remove(_db_file)

    _manage = os.path.join(install_dir, 'manage.py')

    _sub_cmds = [
        ('makemigrations', 'custom_user'),
        ('makemigrations', 'data_management'),
        ('migrate',),
        (
            'graph_models',
            'data_management',
            '--arrow-shape',
            'crow',
            '-x',
            '"BaseModel,DataObject,DataObjectVersion"',
            '-E',
            '-o',
            os.path.join(install_dir, 'schema.dot')
        ),
        ('collectstatic', '--noinput'),
        ('createsuperuser', '--noinput')
    ]

    logger.debug("Running Django migration commands as setup")

    for sub in _sub_cmds:
        subprocess.check_call(
            [python, _manage, *sub],
            shell=False,
            stdout=subprocess.DEVNULL if silent else None,
            env=django_environ()
        )

    if shutil.which('dot'):
        logger.debug("Creating schema diagram")
        subprocess.check_call(
            [
                shutil.which('dot'),
                os.path.join(install_dir, 'schema.dot'),
                '-Tsvg',
                '-o',
                os.path.join(install_dir, 'static', 'images', 'schema.svg')
            ],
            shell=False,
            stdout=subprocess.DEVNULL if silent else None
        )


def install_registry(
    repository: str = FAIR_REGISTRY_REPO,
    reference: str = None,
    install_dir: str = None,
    silent: bool = False,
    force: bool = False,
    venv_dir: str = None) -> None:

    logger = logging.getLogger('FAIRDataPipeline.Server')

    if not install_dir:
        install_dir = DEFAULT_REGISTRY_LOCATION

    if os.path.exists(install_dir):
        raise fdp_exc.RegistryError(f"Local registry is already installed in {install_dir}")

    logger.debug("Installing registry to '%s'", install_dir)
    
    # In case registry installation occurs before setup
    os.makedirs(fdp_com.global_config_dir(), exist_ok=True) 

    logger.debug("Updating registry path in global CLI configuration")
    if os.path.exists(fdp_com.global_fdpconfig()):
        _glob_conf = fdp_util.flatten_dict(fdp_conf.read_global_fdpconfig())
    else:
        _glob_conf = {}

    _glob_conf['registries.local.directory'] = install_dir

    with open(fdp_com.global_fdpconfig(), 'w') as out_conf:
        yaml.dump(fdp_util.expand_dict(_glob_conf), out_conf)
    
    if force:
        logger.debug("Removing existing installation at '%s'", install_dir)
        shutil.rmtree(install_dir, ignore_errors=True)

    logger.debug("Creating directories for installation if they do not exist")

    os.makedirs(os.path.dirname(install_dir), exist_ok=True)

    _repo = git.Repo.clone_from(repository, install_dir)

    logger.debug("Retrieving the reference specified from the repository")

    # If no reference is specified, use the latest tag for the registry
    if not reference:
        if not _repo.tags:
            raise fdp_exc.RegistryError(
                f"Expected tags in local registry repository '{repository}', "
                "but found none"
            )
        reference = _repo.tags[-1].name

    logger.debug("Using reference '%s' for registry checkout", reference)

    _candidates = [t.name for t in _repo.tags+_repo.heads]

    if reference not in _candidates:
        raise fdp_exc.RegistryError(
            f"No such head or tag '{reference}' in registry repository"
            f"Candidates are {_candidates}"
        )
    else:
        _repo.git.checkout(reference)

    if not venv_dir:
        venv_dir = os.path.join(install_dir, 'venv')

        venv.create(venv_dir, with_pip=True, prompt='TestRegistry',)

    _python_exe = 'python.exe' if platform.system() == 'Windows' else 'python'
    _binary_loc = 'Scripts' if platform.system() == 'Windows' else 'bin'

    _venv_python = shutil.which(_python_exe, path=os.path.join(venv_dir, _binary_loc))

    if not _venv_python:
        raise FileNotFoundError(
            f"Failed to find binary '{_python_exe}' in location '{venv_dir}"
        )

    logger.debug("Installing requirements")

    subprocess.check_call(
        [_venv_python, '-m', 'pip', 'install', '--upgrade', 'pip', 'wheel'],
        shell=False,
        stdout=subprocess.DEVNULL if silent else None
    )

    subprocess.check_call(
        [_venv_python, '-m', 'pip', 'install', 'whitenoise'],
        shell=False,
        stdout=subprocess.DEVNULL if silent else None
    )

    _requirements = os.path.join(install_dir, 'local-requirements.txt')

    if not os.path.exists(_requirements):
        raise FileNotFoundError(
            f"Expected file '{_requirements}'"
        )

    subprocess.check_call(
        [_venv_python, '-m', 'pip', 'install', '-r', _requirements],
        shell=False,
        stdout=subprocess.DEVNULL if silent else None
    )

    rebuild_local(_venv_python, install_dir, silent)


def uninstall_registry() -> None:
    """Uninstall the local registry from the default location"""
    logger = logging.getLogger('FAIRDataPipeline.Server')
    # First check if the location can be retrieved from a CLI configuration
    # else check the default install location
    if (os.path.exists(fdp_com.global_fdpconfig())
        and os.path.exists(fdp_com.registry_home())):
        logger.debug("Uninstalling registry, removing '%s'", fdp_com.registry_home())
        shutil.rmtree(fdp_com.registry_home())
    elif os.path.exists(DEFAULT_REGISTRY_LOCATION):
        logger.debug("Uninstalling registry, removing '%s'", DEFAULT_REGISTRY_LOCATION)
        shutil.rmtree(DEFAULT_REGISTRY_LOCATION)
    else:
        raise fdp_exc.RegistryError(
            "Cannot uninstall registry, no local installation identified"
        )


def update_registry_post_setup(repo_dir: str, global_setup: bool = False) -> None:
    """Add user namespace and file types after CLI setup

    Parameters
    ----------
    repo_dir : str
        current FAIR repository location
    global_setup : bool, optional
        whether this is the first (global) setup or local for a new repository
    """
    logger = logging.getLogger('FAIRDataPipeline.Server')

    logger.debug("Updating registry from latest setup")

    _is_running = check_server_running(fdp_conf.get_local_uri())

    if not _is_running:
        launch_server(fdp_conf.get_local_uri())

    if global_setup:
        logger.debug("Populating file types")
        fdp_store.populate_file_type(fdp_conf.get_local_uri())

    logger.debug("Adding 'author' and 'UserAuthor' entries ig not present")
    # Add author and UserAuthor
    _author_url = fdp_store.store_user(repo_dir, fdp_conf.get_local_uri())

    try:
        _admin_url = fdp_req.get(
            fdp_conf.get_local_uri(),
            'users',
            params = {
                'username': 'admin'
            }
        )[0]['url']
    except (KeyError, IndexError):
        raise fdp_exc.RegistryAPICallError(
            "Failed to retrieve 'admin' user from registry database"
        )

    fdp_req.post_else_get(
        fdp_conf.get_local_uri(),
        'user_author',
        data = {
            'user': _admin_url,
            'author': _author_url
        }
    )

    # Only stop the server if it was not running initially
    if not _is_running:
        stop_server()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Identifiers
===========

Methods relating to retrieval and parsing of unique identifiers

Contents
========

Functions
---------
    check_orcid - verify an ORCID and return name information
    check_ror - verify a ROR and return name information
    check_grid - verify a GRID ID and return name information
"""

__date__ = "2021-07-01"

import urllib.parse
import typing
import time
import requests
import requests.exceptions

ID_URIS = {
    'orcid': 'https://orcid.org/',
    'ror': 'https://ror.org/',
    'grid': 'https://www.grid.ac/institutes/'
}

QUERY_URLS = {
    'orcid': "https://pub.orcid.org/v2.0/",
    'ror': "https://api.ror.org/organizations?query=",
    'grid': "https://www.grid.ac/institutes/"
}


def check_orcid(orcid: str) -> typing.Dict:
    """Checks if valid ORCID using ORCID public api

    Parameters
    ----------
    orcid : str
        ORCID to be checked

    Returns
    -------
    typing.Dict
        metadata from the given ID
    """

    _header = {'Accept': 'application/json'}
    _url = urllib.parse.urljoin(QUERY_URLS['orcid'], orcid)
    _response = requests.get(_url, headers = _header)

    _result_dict: typing.Dict[str, typing.Any] = {}

    if _response.status_code != 200:
        return _result_dict
    
    _names = _response.json()['person']['name']
    _given = _names['given-names']['value']
    _family = _names['family-name']['value']
    _name = _given + ' ' + _family

    _result_dict['name'] = _name
    _result_dict['family_name'] = _family
    _result_dict['given_names'] = _given
    _result_dict['orcid'] = orcid
    _result_dict['uri'] = f'{ID_URIS["orcid"]}{orcid}'

    return _result_dict

def check_ror(ror: str) -> typing.Dict:
    """Checks if valid ROR using ROR public api

    Parameters
    ----------
    ror : str
        ROR to be checked

    Returns
    -------
    typing.Dict
        metadata from the given ID
    """

    _url = f"{QUERY_URLS['ror']}{ror}"
    _response = requests.get(_url)

    _result_dict: typing.Dict[str, typing.Any] = {}

    if _response.status_code != 200:
        return _result_dict
    
    if _response.json()['number_of_results'] == 0:
        return _result_dict
        
    _name = _response.json()['items'][0]['name']
    _result_dict['name'] = _name
    _result_dict['family_name'] = _name
    _result_dict['given_names'] = None
    _result_dict['ror'] = ror
    _result_dict['uri'] = f'{ID_URIS["ror"]}{ror}'

    return _result_dict


def check_grid(grid_id: str) -> typing.Dict:
    """Checks if valid GRID ID using GRID public api
    Parameters
    ----------
    grid_id : str
        GRID ID to be checked
    Returns
    -------
    typing.Dict
        metadata from the given ID
    """
    _header = {'Accept': 'application/json'}
    _response = requests.get(f'{QUERY_URLS["grid"]}{grid_id}', headers=_header)

    _result_dict: typing.Dict[str, typing.Any] = {}

    if _response.status_code != 200:
        return _result_dict

    _name = _response.json()['institute']['name']

    _result_dict['name'] = _name
    _result_dict['family_name'] = _name
    _result_dict['given_names'] = None
    _result_dict['grid'] = grid_id
    _result_dict['uri'] = f'{ID_URIS["grid"]}{grid_id}'

    return _result_dict


def check_id_permitted(identifier: str, retries: int = 5) -> bool:
    """Check a user provided identifier is permitted

    This ID is expected to be a valid URL

    Parameters
    ----------
    identifier : str
        identifier URL candidate

    Returns
    -------
    bool
        if valid identifier
    """
    _n_attempts = 0

    while _n_attempts < retries:
        try:
            requests.get(identifier).raise_for_status()
            return True
        except (
            requests.exceptions.MissingSchema,
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError):
            _n_attempts += 1
            time.sleep(1)
            continue
    
    return False



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Parse User Config
=================

Perform parsing of the user updated `config.yaml` file.


Contents
========

Modules
-------

    globbing - swap glob expressions for registry entries
    variables  - substitute recognised FAIR CLI variables 

"""

__date__ = "2021-08-04"







#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
FAIR-CLI Exceptions
===================

Custom exceptions for the command line interface. These are captured during
command line usage, but provide a means examining scenarios during testing.


Contents
========

Exceptions
----------

    FAIRCLIException
    RegistryErrorException
    CLIConfigurationError
    KeyPathError
    UnexpectedRegistryServerState
    FDPRepositoryError
    FileNotFoundError
    InternalError
    StagingError

"""

__date__ = "2021-06-28"

import click


class FAIRCLIException(Exception):
    """Base exception class for all FAIR-CLI exceptions"""

    def __init__(
        self,
        msg: str,
        hint: str = "",
        level: str = "Error",
        exit_code: int = 1,
    ):
        """Initialises a FAIR-CLI exception type.

        A level can be specified which is a prefix whenever the exception is
        captured and printed (when used within the CLI itself)

        Parameters
        ----------
        msg : str
            message to display
        hint : str
            possible solution to issue raised
        level : str, optional
            level descriptor (message prefix), by default "Error"
        """
        self.msg = msg
        self.level = level
        self.hint = hint
        self.exit_code = exit_code
        super().__init__(msg)

    def err_print(self) -> None:
        _out_msg = f"{self.level+': ' if self.level else ''}{self.msg}"
        if self.hint:
            _out_msg += f"\n{self.hint}"
        click.echo(_out_msg)


class RegistryError(FAIRCLIException):
    """Errors relating to registry setup and usage"""

    def __init__(self, msg: str, hint:str = ""):
        super().__init__(msg, hint=hint)


class CLIConfigurationError(FAIRCLIException):
    """Errors relating to CLI configuration"""

    def __init__(self, msg, hint="", level='Error'):
        super().__init__(msg, hint=hint, level=level)


class KeyPathError(FAIRCLIException):
    """Errors relating to key path within a nested mapping"""

    def __init__(self, key, parent_label):
        _msg = f"Failed to retrieve item at address '{key}' from mapping '{parent_label}', no such address"
        super().__init__(_msg, level='Error')


class UserConfigError(FAIRCLIException):
    """Errors relating to the user 'config.yaml' file"""

    def __init__(self, msg, hint=""):
        super().__init__(msg, hint=hint)


class UnexpectedRegistryServerState(FAIRCLIException):
    """Errors relating to server start/stop when already up/down)"""

    def __init__(self, msg, hint=""):
        super().__init__(msg, hint)


class FDPRepositoryError(FAIRCLIException):
    """Errors relating to FAIR repository"""

    def __init__(self, msg, hint=""):
        super().__init__(msg, hint=hint)


class FileNotFoundError(FAIRCLIException):
    """Attempt to access a non-existent file"""

    def __init__(self, msg, hint=""):
        super().__init__(msg, hint=hint)


class InternalError(FAIRCLIException):
    """Errors relating to non-user created issues"""

    def __init__(self, msg):
        super().__init__(msg, level="InternalError")


class CommandExecutionError(FAIRCLIException):
    """Errors related to execution of submission script commands"""

    def __init__(self, msg, exit_code):
        super().__init__(msg, exit_code=exit_code)


class RegistryAPICallError(FAIRCLIException):
    """Errors relating to invalid queries to the registry RestAPI"""

    def __init__(self, msg, error_code):
        self.error_code = error_code
        _level = "Warning" if self.error_code in [403] else "Error"
        super().__init__(f'[HTTP {self.error_code}]: {msg}', exit_code=error_code, level=_level)

class NotImplementedError(FAIRCLIException):
    """Errors relating to features that have not yet been implemented"""
    def __init__(self, msg, hint="", level='Error'):
        super().__init__(msg, hint, level=level)


class StagingError(FAIRCLIException):
    """Errors relating to the staging of jobs"""
    def __init__(self, msg):
        super().__init__(msg)


class SynchronisationError(FAIRCLIException):
    """Errors relating to synchronisation between registries"""
    def __init__(self, msg, error_code):
        self.error_code = error_code
        super().__init__(msg, exit_code=error_code)


class ImplementationError(FAIRCLIException):
    """Errors relating to setup via API implementation"""
    def __init__(self, msg):
        super().__init__(msg)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Registry Sync
=============

Methods for synchronising across registries and allocating versioning

Contents
========

Functions
---------

"""

__date__ = "2021-08-05"

import typing
import os
import collections
import urllib.parse
import logging
import click
import yaml

import fair.exceptions as fdp_exc
import fair.registry.requests as fdp_req

_logger = logging.getLogger('FairDataPipeline.Sync')


def get_dependency_chain(object_url: str) -> collections.deque:
    """Get all objects relating to an object in order of dependency

    For a given URL this function fetches all component URLs ordering them
    in terms of creation order (no dependencies -> most dependencies)
    allowing them to then be re-created in the correct order.

    Parameters
    ----------
    object_url : str
        Full URL of an object within a registry

    Returns
    -------
    collections.deque
        ordered iterable of component object URLs
    """
    _logger.debug(f"Retrieving dependency chain for '{object_url}'")
    _local_uri, _ = fdp_req.split_api_url(object_url)

    _dependency_list = fdp_req.get_dependency_listing(_local_uri)

    def _dependency_of(url_list: collections.deque, item: str):
        if item in url_list:
            return
        url_list.appendleft(item)
        _results = fdp_req.url_get(item)
        for req in _results:
            if req in _dependency_list and _results[req]:
                _dependency_of(url_list, _results[req])

    # Ordering is important so use a deque to preserve
    _urls = collections.deque()
    _dependency_of(_urls, object_url)

    return _urls


def push_dependency_chain(
    object_url: str,
    dest_uri: str,
    dest_token: str,
    ) -> typing.Dict[str, str]:
    """Push an object and all of its dependencies to the remote registry

    In order to push an object, firstly any dependencies which are also
    registry items must be pushed first.

    Parameters
    ----------
    object_url : str
        object to push
    dest_uri : str
        endpoint of the destination registry
    dest_token : str
        access token for the destination registry

    Returns
    -------
    typing.Dict[str, str]
        dictionary showing conversion from source registry URL to destination
    """
    _logger.debug(
        f"Attempting to push object '{object_url}' to '{dest_uri}'"
    )

    _dependency_chain: collections.deque = get_dependency_chain(object_url)
    _new_urls: typing.Dict[str, str] = {
        k: ""
        for k in _dependency_chain
    }

    for object in _dependency_chain:
        _obj_data = fdp_req.url_get(object)
        _obj_type = fdp_req.get_obj_type_from_url(object)
        _uri, _ = fdp_req.split_api_url(object)
        # Substitute any local dependency URLs for those
        # from the components created on the remote
        for key, value in _obj_data:
            if value in _new_urls:
                _obj_data[key] = _new_urls[value]
        _writable_data = {
            k: v for k, v in _obj_data.items()
            if k in fdp_req.get_writable_fields(_uri, _obj_type)
        }

        # Filters are all variables returned by 'filter_fields' request for a
        # given object minus any variables which have a URL value
        # (as remote URL will never match local)
        _filters = {
            k: v for k, v in _obj_data.items()
            if k in fdp_req.get_filter_variables(_uri, _obj_type) and
            not urllib.parse.urlparse(v).netloc
        }

        _logger.debug(
            f"Pushing member '{object}' to '{dest_uri}'"
        )
        _new_url = fdp_req.post_else_get(
            dest_uri,
            _obj_type,
            data=_writable_data,
            token=dest_token,
            params=_filters
        )

        _new_urls[object] = _new_url

    return _new_urls
        

def push_from_config(
    local_uri: str,
    dest_uri: str,
    dest_token: str,
    config_yaml: str) -> None:
    if not os.path.exists(config_yaml):
        raise fdp_exc.FileNotFoundError(
            f"Cannot load write statements from '{config_yaml}', "
            "file does not exist."
        )
    _logger.debug(f"Reading 'write' statement in '{config_yaml}'")
    _config = yaml.safe_load(open(config_yaml))

    if 'write' not in _config:
        click.echo("Nothing to push.")
        return

    for object in _config['write']:
        _logger.debug(f"Processing object '{object}'")
        if 'external_object' in object:
            _usable_fields = {}
            if 'identifier' in object:
                _usable_fields['identifier'] = object['identifier']
            elif 'alternate_identifier' in object:
                _usable_fields['identifier'] = object['alternate_identifier']
            if 'title' in object:
                _usable_fields['title'] = object['title']
            if 'version' in object:
                _usable_fields['version'] = object['version']
            _entries = fdp_req.get(
                local_uri,
                'external_object',
                params=_usable_fields
            )

            if not _entries or len(_entries) > 1:
                raise fdp_exc.InternalError(
                    "Expected single entry for 'external_object' "
                    f"'{object['external_object']}"
                )
            
            _url = _entries[0]['url']

            push_dependency_chain(_url, dest_uri, dest_token)
        elif 'data_product' in object:
            _usable_fields = {'name': object['data_product']}
            if 'version' in object:
                _usable_fields['version'] = object['version']
            
            _entries = fdp_req.get(
                local_uri,
                'data_product',
                params=_usable_fields
            )

            if not _entries or len(_entries) > 1:
                raise fdp_exc.InternalError(
                    "Expected single entry for 'data_product' "
                    f"'{object['data_product']}"
                )
            
            _url = _entries[0]['url']

            push_dependency_chain(_url, dest_uri, dest_token)
        else:
            fdp_exc.NotImplementedError(
                "Cannot write unsupported data type, object must be "
                "either a data_product or external_object"
            )
            


#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Session
=======

Manage synchronisation of data and metadata relating to runs using the
FAIR Data Pipeline system.

Contents
========

Classes
-------

    FAIR - main class for performing synchronisations and executing jobs

Misc Variables
--------------

    __author__
    __license__
    __credits__
    __status__
    __copyright__

"""

__date__ = "2021-06-28"

import os
import glob
import uuid
import typing
import pathlib
import logging
import shutil
import copy
import click
import rich
import git
import yaml

import fair.templates as fdp_tpl
import fair.common as fdp_com
import fair.run as fdp_run
import fair.registry.server as fdp_serv
import fair.configuration as fdp_conf
import fair.exceptions as fdp_exc
import fair.history as fdp_hist
import fair.staging as fdp_stage
import fair.testing as fdp_test
import fair.registry.sync as fdp_sync


class FAIR:
    """
    A class which provides the main interface for managing runs and data
    transfer between FAIR Data Pipeline registries.

    Methods are based around a user directory which is specified with locations
    being determined relative to the closest FAIR repository root folder (i.e.
    the closest location in the upper hierarchy containing a '.fair' folder).
    """
    _logger = logging.getLogger("FAIRDataPipeline.Session")

    def __init__(
        self,
        repo_loc: str,
        user_config: str = None,
        debug: bool = False,
        server_mode: fdp_serv.SwitchMode = fdp_serv.SwitchMode.NO_SERVER,
        testing: bool = False
    ) -> None:
        """Initialise instance of FAIR sync tool

        All actions are performed relative to the specified folder after the
        local '.fair' directory for the repository has been located.

        A session is usually cached to ensure that the server is not shut down
        when a process is taking place. The exception is where the user
        explicitly requests for it to be started.

        Parameters
        ----------
        repo_loc : str
            location of FAIR repository to update
        user_config : str, optional
            alternative config.yaml user configuration file
        debug : bool, optional
            run in verbose mode
        server_mode : fair.registry.server.SwitchMode, optional
            stop/start server mode during session
        testing : bool
            run in testing mode
        """
        if debug:
            logging.getLogger('FAIRDataPipeline').setLevel(logging.DEBUG)
        self._logger.debug("Starting new session.")
        self._testing = testing
        self._session_loc = repo_loc
        self._logger.debug(f"Session location: {self._session_loc}")
        self._run_mode = server_mode
        self._stager: fdp_stage.Stager = fdp_stage.Stager(self._session_loc)
        self._session_id = (
            uuid.uuid4() if server_mode == fdp_serv.SwitchMode.CLI else None
        )
        self._session_config = user_config or fdp_com.local_user_config(
            self._session_loc
        )

        if server_mode != fdp_serv.SwitchMode.NO_SERVER and not os.path.exists(
            fdp_com.registry_home()
        ):
            raise fdp_exc.RegistryError(
                "User registry directory was not found, this could "
                "mean the local registry has not been installed."
            )

        if not os.path.exists(fdp_com.global_config_dir()):
            self._logger.debug(
                "Creating directory: %s", fdp_com.global_config_dir()
            )
            os.makedirs(fdp_com.global_config_dir())

        # Initialise all configuration status dictionaries
        self._local_config: typing.Dict[str, typing.Any] = {}
        self._global_config: typing.Dict[str, typing.Any] = {}

        self._logger.debug(
            "Initialising session with:\n"
            "\tsession_config = %s\n"
            "\ttesting        = %s\n"
            "\trun_mode       = %s\n"
            "\tstaging_file   = %s\n"
            "\tsession_id     = %s\n",
            self._session_config,
            self._testing,
            self._run_mode,
            self._stager._staging_file,
            self._session_id
        )

        self._load_configurations()

        self._setup_server()

    def push(self, remote: str = 'origin'):
        self._logger.debug(
            f"Pushing items in '{self._session_config}:write' to remote"
            f" registry '{remote}'"
        )
        fdp_sync.push_from_config(
            fdp_conf.get_local_uri(),
            fdp_conf.get_remote_uri(self._session_loc, remote),
            fdp_conf.get_remote_token(self._session_loc, remote),
            self._session_config
        )

    def purge(
        self,
        verbose: bool = True,
        local_cfg: bool = False,
        global_cfg: bool = False,
        clear_data: bool = False,
        clear_all: bool = False
        ) -> None:
        """Remove FAIR-CLI tracking from the given directory

        Parameters
        ==========
        local_cfg : bool, optional
            remove local directories, default is False
        global_cfg : bool, optional
            remove global directories, default is False
        verbose : bool, optional
            run in verbose mode, default is True
        clear_data : bool, optional
            remove the data directory (potentially dangerous), default is False
        clear_all : bool, optional
            remove all FAIR components from the system, overrides others, default is False
        """
        _root_dir = os.path.join(fdp_com.find_fair_root(self._session_loc), fdp_com.FAIR_FOLDER)
        if os.path.exists(_root_dir):
            if verbose:
                click.echo(f"Removing directory '{_root_dir}'")
            shutil.rmtree(_root_dir)
        if clear_all:
            if verbose:
                click.echo(f"Removing directory '{fdp_com.USER_FAIR_DIR}'")
            shutil.rmtree(fdp_com.USER_FAIR_DIR)
            return
        if clear_data:
            try:
                if verbose:
                    click.echo(
                        f"Removing directory '{fdp_com.default_data_dir()}'"
                    )
                if os.path.exists(fdp_com.default_data_dir()):
                    shutil.rmtree(fdp_com.default_data_dir())
            except FileNotFoundError:
                raise fdp_exc.FileNotFoundError(
                    "Cannot remove local data store, a global CLI configuration "
                    "is required to identify its location"
                )
        if global_cfg:
            if verbose:
                click.echo(f"Removing directory '{fdp_com.global_config_dir()}'")
            _global_dirs = fdp_com.global_config_dir()
            if os.path.exists(_global_dirs):
                shutil.rmtree(_global_dirs)

    def _setup_server(self) -> None:
        """Start or stop the server if required"""

        # If a session ID has been specified this means the server is auto
        # started as opposed to being started explcitly by the user
        # this means it will be shut down on completion
        self._logger.debug(f"Running server setup for run mode {self._run_mode}")
        if self._run_mode == fdp_serv.SwitchMode.CLI:
            self._setup_server_cli_mode()
        elif self._run_mode == fdp_serv.SwitchMode.USER_START:
            self._setup_server_user_start()
        elif self._run_mode in [
            fdp_serv.SwitchMode.USER_STOP,
            fdp_serv.SwitchMode.FORCE_STOP,
        ]:
            _cache_addr = os.path.join(fdp_com.session_cache_dir(), 'user.run')
            if not fdp_serv.check_server_running():
                raise fdp_exc.UnexpectedRegistryServerState(
                    "Server is not running."
                )
            if os.path.exists(_cache_addr):
                os.remove(_cache_addr)
            click.echo("Stopping local registry server.")
            fdp_serv.stop_server(
                force=self._run_mode == fdp_serv.SwitchMode.FORCE_STOP,
            )

    def _setup_server_cli_mode(self):
        self.check_is_repo()
        _cache_addr = os.path.join(
            fdp_com.session_cache_dir(), f"{self._session_id}.run"
        )

        self._logger.debug("Checking for existing sessions")
        # If there are no session cache files start the server
        if not glob.glob(
            os.path.join(fdp_com.session_cache_dir(), "*.run")
        ):
            self._logger.debug("No sessions found, launching server")
            fdp_serv.launch_server()

        self._logger.debug(f"Creating new session #{self._session_id}")

        if not os.path.exists(fdp_com.session_cache_dir()):
            raise fdp_exc.InternalError(
                "Failed to create session cache file, "
                f"expected cache directory '{fdp_com.session_cache_dir()}' to exist"
            )

        # Create new session cache file
        pathlib.Path(_cache_addr).touch()

    def _setup_server_user_start(self):
        if not os.path.exists(fdp_com.session_cache_dir()):
            os.makedirs(fdp_com.session_cache_dir())

        _cache_addr = os.path.join(fdp_com.session_cache_dir(), 'user.run')

        if "registries" not in self._global_config:
            raise fdp_exc.CLIConfigurationError(
                "Cannot find server address in current configuration",
                hint="Is the current location a FAIR repository?"
            )

        if fdp_serv.check_server_running():
            raise fdp_exc.UnexpectedRegistryServerState(
                "Server already running."
            )
        click.echo("Starting local registry server")
        pathlib.Path(_cache_addr).touch()
        fdp_serv.launch_server(fdp_conf.get_local_uri(), verbose=True)

    def run_job(
        self,
        bash_cmd: str = "",
        mode: fdp_run.CMD_MODE = fdp_run.CMD_MODE.RUN
    ) -> str:
        """Execute a run using the given user configuration file"""
        self.check_is_repo()
        if not os.path.exists(self._session_config):
            self.make_starter_config()
        
        self._logger.debug("Setting up command execution")

        _hash = fdp_run.run_command(
            repo_dir=self._session_loc,
            config_yaml=self._session_config,
            bash_cmd=bash_cmd,
            mode=mode
        )

        self._logger.debug(f"Tracking job hash {_hash}")

        # Automatically add the run to tracking but unstaged
        self._stager.change_job_stage_status(_hash, False)

        return _hash

    def check_is_repo(self, location: str = None) -> None:
        """Check that the current location is a FAIR repository"""
        if not location:
            location = self._session_loc
        if not fdp_com.find_fair_root(location):
            raise fdp_exc.FDPRepositoryError(
                f"'{location}' is not a FAIR repository", hint="Run 'fair init' to initialise."
            )

    def check_git_repo_state(self, git_repo: str, remote_label: str = 'origin') -> bool:
        """Checks the git repository is clean and that local matches remote"""
        _repo_root = fdp_com.find_git_root(git_repo)
        _repo = git.Repo(_repo_root)

        # Firstly get the current branch
        _current_branch = _repo.active_branch.name

        # Get the latest commit on the current branch locally
        _loc_commit = _repo.refs[_current_branch].commit.hexsha

        # Get the latest commit on this branch on remote
        _rem_commit = _repo.remotes[remote_label].refs[_current_branch].commit.hexsha

        # Commit match
        _com_match = _loc_commit == _rem_commit

        return _repo.is_dirty() and _com_match

    def __enter__(self) -> None:
        """Method called when using 'with' statement."""
        return self

    def _load_configurations(self) -> None:
        """This ensures all configurations are read at the
        start of every session.
        """
        self._logger.debug("Loading CLI configurations.")

        if os.path.exists(fdp_com.global_fdpconfig()):
            self._global_config = fdp_conf.read_global_fdpconfig()
        if os.path.exists(fdp_com.local_fdpconfig(self._session_loc)):
            self._local_config = fdp_conf.read_local_fdpconfig(
                self._session_loc
            )

    def change_staging_state(
        self, job_to_stage: str, stage: bool = True
    ) -> None:
        """Change the staging status of a given run

        Parameters
        ----------
        job_to_stage : str
            uuid of run to add to staging
        stage : bool, optional
            whether to stage/unstage run, by default True (staged)
        """
        self.check_is_repo()
        self._stager.change_job_stage_status(job_to_stage, stage)

    def remove_job(self, job_id: str, cached: bool = False) -> None:
        """Remove a job from tracking

        Parameters
        ----------
        file_name : str
            path of file to be removed
        cached : bool, optional
            remove from tracking but not from system, by default False
        """
        self.check_is_repo()

        self._stager.remove_staging_entry(job_id)

    def add_remote(
        self,
        remote_url: str,
        token_file: str,
        label: str = "origin"
        ) -> None:
        """Add a remote to the list of remote URLs"""
        self.check_is_repo()
        if "registries" not in self._local_config:
            self._local_config['registries'] = {}
        if label in self._local_config['registries']:
            raise fdp_exc.CLIConfigurationError(
                f"Registry remote '{label}' already exists."
            )
        self._local_config['registries'][label] = {
            'uri': remote_url,
            'token': token_file
        }

    def remove_remote(self, label: str) -> None:
        """Remove a remote URL from the list of remotes by label"""
        self.check_is_repo()
        if (
            "registries" not in self._local_config
            or label not in self._local_config
        ):
            raise fdp_exc.CLIConfigurationError(
                f"No such entry '{label}' in available remotes"
            )
        del self._local_config[label]

    def modify_remote(self, label: str, url: str) -> None:
        """Update a remote URL for a given remote"""
        self.check_is_repo()
        if (
            "registries" not in self._local_config
            or label not in self._local_config['registries']
        ):
            raise fdp_exc.CLIConfigurationError(
                f"No such entry '{label}' in available remotes"
            )
        self._local_config['registries'][label]['uri'] = url

    def clear_logs(self) -> None:
        """Delete all local run stdout logs

        This does NOT delete any information from the registry
        """
        _log_files = glob.glob(
            fdp_hist.history_directory(self._session_loc), "*.log"
        )
        if _log_files:
            for log in _log_files:
                os.remove(log)

    def list_remotes(self, verbose: bool = False) -> typing.List[str]:
        """List the available RestAPI URLs"""
        self.check_is_repo()
        if "registries" not in self._local_config:
            return []
        _remote_print = []
        for remote, data in self._local_config['registries'].items():
            _out_str = f"[bold white]{remote}[/bold white]"
            if verbose:
                _out_str += f"\t[yellow]{data['uri']}[/yellow]"
            _remote_print.append(_out_str)
        rich.print("\n".join(_remote_print))
        return _remote_print

    def status(self, verbose: bool = False) -> None:
        """Get the status of staging"""
        self._logger.debug("Updating staging status")
        self.check_is_repo()

        _staged_jobs = self._stager.get_item_list(True, "job")
        _unstaged_jobs = self._stager.get_item_list(False, "job")

        if _staged_jobs:
            click.echo("Changes to be synchronized:")
            click.echo("\tJobs:")
            for job in _staged_jobs:
                click.echo(click.style(f"\t\t{job}", fg="green"))
                _job_urls = self._stager.get_job_data(
                    fdp_conf.get_local_uri(), job
                )
                if not verbose:
                    continue

                for key, value in _job_urls.items():
                    if not value:
                        continue
                    click.echo(click.style(f"\t\t\t{key}:", fg="green"))
                    if isinstance(value, list):
                        for url in value:
                            click.echo(
                                click.style(f"\t\t\t\t{url}", fg="green")
                            )
                    else:
                        click.echo(
                            click.style(f"\t\t\t\t{value}", fg="green")
                        )

        if _unstaged_jobs:
            click.echo("Changes not staged for synchronization:")
            click.echo('\t(use "fair add <job>..." to stage jobs)')

            click.echo("\tJobs:")

            for job in _unstaged_jobs:
                click.echo(click.style(f"\t\t{job}", fg="red"))
                _job_urls = self._stager.get_job_data(
                    fdp_conf.get_local_uri(),
                    job
                )

                if not verbose:
                    continue

                for key, value in _job_urls.items():
                    if not value:
                        continue
                    click.echo(
                        click.style(
                            f"\t\t\t{key.replace('_', ' ').title()}:",
                            fg="red"
                        )
                    )
                    if isinstance(value, list):
                        for url in value:
                            click.echo(
                                click.style(f"\t\t\t\t{url}", fg="red")
                            )
                    else:
                        click.echo(
                            click.style(f"\t\t\t\t{value}", fg="red")
                        )

        if not _unstaged_jobs and not _staged_jobs:
            click.echo("Nothing marked for tracking.")

    def make_starter_config(self, output_file_name: str = None) -> None:
        """Create a starter config.yaml"""
        if not output_file_name:
            output_file_name = os.path.join(self._session_loc, fdp_com.USER_CONFIG_FILE)
        if os.path.exists(output_file_name):
            click.echo(
                f"The user configuration file '{os.path.abspath(output_file_name)}'"
                " already exists, skipping creation."
            )
            return
        if "registries" not in self._local_config:
            raise fdp_exc.CLIConfigurationError(
                "Cannot generate user 'config.yaml'",
                hint="You need to set the remote URL"
                " by running: \n\n\tfair remote add <url>\n",
            )

        with open(output_file_name, "w") as f:
            _yaml_str = fdp_tpl.config_template.render(
                instance=self,
                data_dir=fdp_com.default_data_dir(),
                local_repo=os.path.abspath(fdp_com.find_fair_root(self._session_loc)),
            )
            _yaml_dict = yaml.safe_load(_yaml_str)

            yaml.dump(_yaml_dict, f, sort_keys = False)

    def _export_cli_configuration(self, output_file: str) -> None:
        _cli_config = fdp_conf.read_global_fdpconfig()
        _loc_config = fdp_conf.read_local_fdpconfig(self._session_loc)
        _cli_config['git'] = _loc_config['git']
        _cli_config['registries'].update(_loc_config['registries'])
        _cli_config['user'].update(_loc_config['user'])
        with open(output_file, 'w') as f:
            yaml.dump(_cli_config, f)

    def initialise(
        self,
        using: typing.Dict = None,
        registry: str = None,
        export_as: str = None
    ) -> None:
        """Initialise an fair repository within the current location

        Parameters
        ----------
        using : str
            load from an existing global CLI configuration file
        """
        _fair_dir = os.path.abspath(
            os.path.join(self._session_loc, fdp_com.FAIR_FOLDER)
        )

        _first_time = not os.path.exists(fdp_com.global_fdpconfig())

        if self._testing:
            using = fdp_test.create_configurations(
                registry,
                fdp_com.find_git_root(os.getcwd()),
                os.getcwd()
            )

        if os.path.exists(_fair_dir):
            if export_as:
                self._export_cli_configuration(export_as)
                return
            click.echo('FAIR repository is already initialised.')
            return

        if not using:
            click.echo(
                "Initialising FAIR repository, setup will now ask for basic info:\n"
            )

        if not os.path.exists(_fair_dir):
            os.mkdir(_fair_dir)
            os.makedirs(fdp_com.session_cache_dir(), exist_ok=True)
            if using:
                self._validate_and_load_cli_config(using)
            self._stager.initialise()

        if not os.path.exists(fdp_com.global_fdpconfig()):
            try:
                self._global_config = fdp_conf.global_config_query(
                    registry
                )
            except (fdp_exc.CLIConfigurationError, click.Abort) as e:
                self._clean_reset(_fair_dir, e)
            try:
                self._local_config = fdp_conf.local_config_query(
                        self._global_config, first_time_setup=_first_time
                    )
            except (fdp_exc.CLIConfigurationError, click.Abort) as e:
                self._clean_reset(_fair_dir, e, True)
        elif not using:
            try:
                self._local_config = fdp_conf.local_config_query(
                    self._global_config
                )
            except (fdp_exc.CLIConfigurationError, click.Abort) as e:
                self._clean_reset(_fair_dir, e, True)
        if not using:
            with open(fdp_com.local_fdpconfig(self._session_loc), "w") as f:
                yaml.dump(self._local_config, f)
            with open(fdp_com.global_fdpconfig(), "w") as f:
                yaml.dump(self._global_config, f)
        else:
            self._global_config = fdp_conf.read_global_fdpconfig()
            self._local_config = fdp_conf.read_local_fdpconfig(self._session_loc)

        if export_as:
            self._export_cli_configuration(export_as)

        fdp_serv.update_registry_post_setup(self._session_loc, _first_time)

        click.echo(f"Initialised empty fair repository in {_fair_dir}")

    def _clean_reset(self, _fair_dir, e, local_only: bool = False):
        if not local_only:
            shutil.rmtree(fdp_com.session_cache_dir(), ignore_errors=True)
            shutil.rmtree(fdp_com.global_config_dir(), ignore_errors=True)
        shutil.rmtree(_fair_dir)
        raise e

    def close_session(self) -> None:
        """Upon exiting, dump all configurations to file"""
        if not os.path.exists(
            os.path.join(self._session_loc, fdp_com.FAIR_FOLDER)
        ):
            return

        if self._session_id:
            # Remove the session cache file
            _cache_addr = os.path.join(
                fdp_com.session_cache_dir(), f"{self._session_id}.run"
            )
            os.remove(_cache_addr)

        if (not os.path.exists(os.path.join(fdp_com.session_cache_dir(), "user.run"))
            and self._run_mode != fdp_serv.SwitchMode.NO_SERVER
        ):
            fdp_serv.stop_server()

        with open(fdp_com.global_fdpconfig(), "w") as f:
            yaml.dump(self._global_config, f)
        with open(fdp_com.local_fdpconfig(self._session_loc), "w") as f:
            yaml.dump(self._local_config, f)

    def _validate_and_load_cli_config(self, cli_config: typing.Dict):
        _exp_keys = [
            'registries',
            'namespaces',
            'user',
            'git'
        ]

        for key in _exp_keys:
            if key not in cli_config:
                self.purge(verbose=False)
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key '{key}' in CLI configuration file"
                )

        for exp_reg in ['local', 'origin']:
            if exp_reg not in cli_config['registries']:
                self.purge(verbose=False)
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key 'registries:{exp_reg}' in CLI configuration file"
                )

        for name, reg in cli_config['registries'].items():
            if 'data_store' not in reg:
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key '{key}' for remote '{name}' "
                    "in CLI configuration"
                )
            if name != 'local' and 'uri' not in reg:
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key 'uri' for remote '{name}' "
                    "in CLI configuration"
                )
            if name != 'local' and 'token' not in reg:
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key 'token' for remote '{name}' "
                    "in CLI configuration"
                )
            if name == 'local' and 'directory' not in reg:
                raise fdp_exc.CLIConfigurationError(
                    "Expected key 'directory' for local registry in CLI configuration"
                )


        _user_keys = ['email', 'family_name', 'given_names', 'orcid', 'uuid']

        for key in _user_keys:
            if key not in cli_config['user']:
                self.purge(verbose=False)
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key 'user:{key}' in CLI configuration file"
                )

        if not cli_config['user']['orcid'] and not cli_config['user']['uuid']:
            raise fdp_exc.CLIConfigurationError(
                "At least one of 'user:orcid' and 'user:uuid' must be provided "
                " in CLI configuration"
            )

        for key in ['local_repo', 'remote']:
            if key not in cli_config['git']:
                self.purge(verbose=False)
                raise fdp_exc.CLIConfigurationError(
                    f"Expected key 'git:{key}' in CLI configuration"
                )

        _glob_cfg = copy.deepcopy(cli_config)
        _loc_cfg = copy.deepcopy(cli_config)
        del _glob_cfg['git']
        if 'description' in _glob_cfg:
            del _glob_cfg['description']
        del _loc_cfg['registries']['local']

        with open(fdp_com.global_fdpconfig(), 'w') as f:
            yaml.dump(_glob_cfg, f)

        with open(fdp_com.local_fdpconfig(self._session_loc), 'w') as f:
            yaml.dump(_loc_cfg, f)

    def __exit__(self, *args) -> None:
        self.close_session()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

FDP-CLI Utilities
=================

Collection of methods used which are not specific to the FDP-CLI and so are
defined as 'utilities'.


Contains
========

Functions
---------

    flatten_dict - convert a nested dictionary into a single level version.
    expand_dict  - expands a single level dictionary to a nested version.
    remove_dictlist_dupes - removes duplicates from list of depth 1 dictionaries

Class
-----
    JSONDateTimeEncoder - for allowing datetime strings to be json parsed

"""

__date__ = "2021-08-04"

import typing
import json
import datetime

def flatten_dict(
    in_dict: typing.Dict,
    separator: str = ".",
    _out_dict: typing.Dict = None,
    _parent_key: str = None,
) -> typing.Dict:
    """Flatten a nested dictionary into a single level variant

    Constructs a new dictionary from a nested dictionary of just one level
    using a separator for the keys.

    Parameters
    ----------
    in_dict : Dict
        input dictionary
    separator : str, optional
        separator to use for keys, by default "."

    Returns
    -------
    Dict
        a flattened version of the input dictionary
    """
    if _out_dict is None:
        _out_dict = {}

    for label, value in in_dict.items():
        new_label = (
            f"{_parent_key}{separator}{label}" if _parent_key else label
        )
        if isinstance(value, dict):
            flatten_dict(
                in_dict=value, _out_dict=_out_dict, _parent_key=new_label
            )
            continue

        _out_dict[new_label] = value

    return _out_dict


def expand_dict(
    in_dict: typing.Dict, separator: str = ".", _out_dict: typing.Dict = None
) -> typing.Dict:
    """Expand a flattened dictionary into a nested dictionary

    Expands a dictionary with a separator in the keys to be a nested
    dictionary.

    Parameters
    ----------
    in_dict : Dict
        input single level dictionary
    separator : str, optional
        key separator, by default "."

    Returns
    -------
    Dict
        nested dictionary representation of the input
    """
    if _out_dict is None:
        _out_dict = {}

    for label, value in in_dict.items():
        if separator not in label:
            _out_dict.update({label: value})
            continue
        key, _components = label.split(separator, 1)
        if key not in _out_dict:
            _out_dict[key] = {}
        expand_dict({_components: value}, separator, _out_dict[key])

    return _out_dict


def remove_dictlist_dupes(
    dicts: typing.List[typing.Dict]) -> typing.List[typing.Dict]:
    """Remove duplicate dictionaries from a list of dictionaries
    
    Note: this will only work with single layer dictionaries!

    Parameters
    ----------
    dicts : List[Dict]
        a list of dictionaries
    
    Returns
    -------
    List[Dict]
        new list without duplicates
    """
    # Convert single layer dictionary to a list of key-value tuples
    _tupleify = [
        [(k, v) for k, v in d.items()]
        for d in dicts
    ]
    
    # Only append unique tuple lists
    _set_tupleify = []
    for t in _tupleify:
        if t not in _set_tupleify:
            _set_tupleify.append(t)

    # Convert the tuple list back to a list of dictionaries
    return [
        {i[0]: i[1] for i in kv}
        for kv in _set_tupleify
    ]


def get_nested_key(search_dict: typing.Dict, key_addr: str, separator: str = '.') -> typing.Any:
    """Retrieve a value from a multi-level dictionary

    Parameters
    ----------
    search_dict : typing.Dict
        dictionary to retrieve from
    key_addr : str
        key address of value to recover

    Returns
    -------
    typing.Any
        value at the specified address
    """
    _flattened_dict = flatten_dict(search_dict, separator=separator)
    return _flattened_dict[key_addr]


class JSONDateTimeEncoder(json.JSONEncoder):
    def default(self, date_time_candidate):
        if isinstance(date_time_candidate, datetime.datetime):
            return str(date_time_candidate)
        else:
            return super().default(date_time_candidate)



import typing
import tempfile
import os
import platform
import pathlib
import git

import fair.identifiers as fdp_id


def create_configurations(
    registry_dir: str,
    local_git_dir: str = None,
    testing_dir: str = tempfile.mkdtemp(),
    tokenless: bool = False) -> typing.Dict:
    """
    Setup CLI for testing

    Running without a token limits the functionality, this should only be used
    when testing without need to access a local registry.
    
    Parameters
    ----------
    registry_dir : str
        directory of the local registry
    testing_dir : str
        working directory for the test
    tokenless : optional, bool
        create a fake token, default False
    
    Returns
    -------
    typing.Dict
        A CLI configuration dictionary that can be loaded for a the CLI session
    """
    _loc_data_store = os.path.join(testing_dir, 'data_store') + os.path.sep
    _proj_dir = os.path.join(testing_dir, 'project')

    if tokenless:
        _token = 't35tt0k3n'
        _token_file = os.path.join(registry_dir, 'token.txt')
        with open(_token_file, 'w') as out_f:
            out_f.write(_token)

    if not local_git_dir:
        local_git_dir = _no_git_setup(_proj_dir)
    os.makedirs(_loc_data_store, exist_ok=True)
    _local_uri = 'http://127.0.0.1:8000/api/'
    _origin_uri = 'http://127.0.0.1:8001/api/'
    if platform.system() == "Windows":
        _local_uri = 'http://127.0.0.1:8000/api/'
        _origin_uri = 'http://127.0.0.1:8001/api/'
    return {
        'namespaces': {'input': 'testing', 'output': 'testing'},
        'registries': {
            'local': {
                'data_store': _loc_data_store,
                'directory': registry_dir,
                'uri': _local_uri
            },
            'origin': {
                'data_store': None,
                'token': os.path.join(registry_dir, 'token.txt'),
                'uri': _origin_uri
            }
        },
        'user': {
            'email': 'test@noreply',
            'family_name': 'Test',
            'given_names': 'Interface',
            'orcid': '000-0000-0000-0000',
            'uri': f'{fdp_id.ID_URIS["orcid"]}000-0000-0000-0000',
            'uuid': '2ddb2358-84bf-43ff-b2aa-3ac7dc3b49f1'
        },
        'git': {
            'local_repo': local_git_dir,
            'remote': 'origin'
        },
    }


def _no_git_setup(_proj_dir):
    _repo = git.Repo.init(_proj_dir)
    _repo.create_remote('origin', url='git@notagit.com/nope')
    result = _proj_dir
    _demo_file = os.path.join(_proj_dir, 'first_file')
    pathlib.Path(_demo_file).touch()
    _repo.index.add(_demo_file)
    _repo.index.commit("First commit of test repository")

    return result
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
FAIR-CLI Templates
==================

Contains methods for assembling templates towards creation of configurations
and user information displays.
"""

__date__ = "2021-06-24"

import os
import jinja2

templates_dir = os.path.dirname(__file__)

config_template = jinja2.Template(
    open(os.path.join(templates_dir, "config.jinja")).read()
)

hist_template = jinja2.Template(
    open(os.path.join(templates_dir, "hist.jinja")).read()
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""
Registry Storage
================

Methods that add objects to the data registry.

Contents
========

Functions
---------

    get_write_storage - retrieve/create root data store
    store_user - create author entry for the current user
    populate_file_type - adds entries for standard file types
    create_file_type - create a file type entry
    store_working_config - create locations/objects for a working config

"""

__date__ = "2021-07-02"

import typing
import os
import hashlib
import urllib.parse
import logging
import yaml

import fair.registry.requests as fdp_req
import fair.exceptions as fdp_exc
import fair.configuration as fdp_conf
import fair.identifiers as fdp_id
import fair.registry.file_types as fdp_file
import fair.registry.versioning as fdp_ver

def get_write_storage(uri: str, write_data_store: str) -> str:
    """Construct storage root if it does not exist

    Parameters
    ----------
    uri : str
        end point of the RestAPI
    write_data_store : str
        path of the write data store

    Returns
    -------
    str
        URI of the created/retrieved storage root

    Raises
    ------
    fdp_exc.UserConfigError
        If 'write_data_store' not present in the working config or global config
    """

    # Convert local file path to a valid data store path
    _write_store_root = f"file://{write_data_store}"
    if _write_store_root[-1] != os.path.sep:
        _write_store_root += os.path.sep

    # Check if the data store already exists by querying for it
    _search_root = fdp_req.get(
        uri, "storage_root", params={"root": _write_store_root}
    )

    # If the data store already exists just return the URI else create it
    # and then do the same
    if _search_root:
        return _search_root[0]["url"]

    _post_data = {"root": _write_store_root, "local": True}
    _storage_root = fdp_req.post(uri, "storage_root", data=_post_data)
    return _storage_root["url"]


def store_user(repo_dir: str, uri: str) -> str:
    """Creates an Author entry for the user if one does not exist

    Parameters
    ----------
    uri : str
        registry RestAPI endpoint

    Returns
    -------
    str
        URI for created author
    """
    _user = fdp_conf.get_current_user_name(repo_dir)
    _data = {'name': ' '.join(_user) if _user[1] else _user[0]}

    try:
        _id = fdp_conf.get_current_user_uri(repo_dir)
        _data['identifier'] = _id
        return fdp_req.post_else_get(
            uri, "author", data=_data, params={"identifier": _id}
        )
    except fdp_exc.CLIConfigurationError:
        _uuid = fdp_conf.get_current_user_uuid(repo_dir)
        _data['uuid'] = _uuid
        return fdp_req.post_else_get(
            uri, "author", data=_data, params={"uuid": _uuid}
        )


def populate_file_type(uri:str) -> typing.List[typing.Dict]:
    """Populates file_type table with common file file_types

    Parameters
    ----------
    uri: str
        registry RestAPI end point
    """
    _type_objs = []

    for _extension in fdp_file.FILE_TYPES:
        # Use post_else_get in case some file types exist already
        _result = create_file_type(uri, _extension)
        _type_objs.append(_result)
    return _type_objs


def create_file_type(uri: str, extension: str) -> str:
    """Creates a new file type on the registry

    Parameters
    ----------
    uri : str
        registry RestAPI end point
    extension : str
        file extension

    Returns
    -------
    str
        URI for created file type
    """
    _name = fdp_file.FILE_TYPES[extension]
    return fdp_req.post_else_get(
        uri, "file_type",
        data={"name": _name, "extension": extension.lower()},
        params={"extension": extension.lower()}
    )


def store_working_config(repo_dir: str, uri: str, work_cfg_yml: str) -> str:
    """Construct a storage location and object for the working config

    Parameters
    ----------
    repo_dir : str
        local FAIR repository
    uri : str
        RestAPI end point
    work_cfg_yml : str
        location of working config yaml

    Returns
    -------
    str
        new URI for the created object

    Raises
    ------
    fair.exceptions.RegistryAPICallError
        if bad status code returned from the registry
    """
    _root_store = get_write_storage(uri, work_cfg_yml)

    _work_cfg = yaml.safe_load(open(work_cfg_yml))
    _work_cfg_data_store = _work_cfg["run_metadata"]["write_data_store"]
    _rel_path = os.path.relpath(work_cfg_yml, _work_cfg_data_store)
    _time_stamp_dir = os.path.basename(os.path.dirname(work_cfg_yml))

    # Construct hash from config contents and time stamp
    # NOTE: You can have the same file stored N times for N job runs
    # hence the use of a timestamp in the hashing
    _hashable = open(work_cfg_yml).read() + _time_stamp_dir

    _hash = hashlib.sha1(_hashable.encode("utf-8")).hexdigest()

    _storage_loc_data = {
        "path": _rel_path,
        "storage_root": _root_store,
        "public": _work_cfg['run_metadata'].get('public', True),
        "hash": _hash,
    }

    try:
        _post_store_loc = fdp_req.post(
            uri,
            "storage_location",
            data=_storage_loc_data
        )
    except fdp_exc.RegistryAPICallError as e:
        if not e.error_code == 409:
            raise e
        else:
            raise fdp_exc.RegistryAPICallError(
                f"Cannot post storage_location "
                f"'{_rel_path}' with hash"
                f" '{_hash}', object already exists",
                error_code=409
            )

    _user = store_user(repo_dir, uri)

    _yaml_type = create_file_type(uri, "yaml")

    _desc = f"Working configuration file for timestamp {_time_stamp_dir}"
    _object_data = {
        "description": _desc,
        "storage_location": _post_store_loc["url"],
        "file_type": _yaml_type,
        "authors": [_user],
    }

    return fdp_req.post_else_get(
        uri, "object", data=_object_data, params={"description": _desc}
    )


def store_working_script(
    repo_dir: str,
    uri: str,
    script_path: str,
    working_config: str
    ) -> str:
    """Construct a storage location and object for the CLI run script

    Parameters
    ----------
    repo_dir : str
        local FAIR repository
    uri : str
        RestAPI end point
    script_path : str
        location of working CLI run script
    data_store : str
        data store path

    Returns
    -------
    str
        new URI for the created object

    Raises
    ------
    fair.exceptions.RegistryAPICallError
        if bad status code returned from the registry
    """
    _work_cfg = yaml.safe_load(open(working_config))
    _root_store = get_write_storage(uri, working_config)
    _data_store = _work_cfg["run_metadata"]["write_data_store"]

    _rel_path = os.path.relpath(script_path, _data_store)

    _time_stamp_dir = os.path.basename(os.path.dirname(working_config))

    # Construct hash from config contents and time
    # NOTE: You can have the same file stored N times for N job runs
    # hence the use of a timestamp in the hashing
    _hashable = open(script_path).read() + _time_stamp_dir

    _hash = hashlib.sha1(_hashable.encode("utf-8")).hexdigest()

    _storage_loc_data = {
        "path": _rel_path,
        "storage_root": _root_store,
        "public": _work_cfg['run_metadata'].get('public', True),
        "hash": _hash,
    }

    try:
        _post_store_loc = fdp_req.post(
            uri,
            "storage_location",
            data=_storage_loc_data
        )
    except fdp_exc.RegistryAPICallError as e:
        if not e.error_code == 409:
            raise e
        else:
            raise fdp_exc.RegistryAPICallError(
                f"Cannot post storage_location "
                f"'{_rel_path}' with hash"
                f" '{_hash}', object already exists",
                error_code=409
            )

    _user = store_user(repo_dir, uri)

    _shell_script_type = create_file_type(
        uri, "sh"
    )

    _time_stamp_dir = os.path.basename(os.path.dirname(script_path))
    _desc = f"Run script for timestamp {_time_stamp_dir}"
    _object_data = {
        "description": _desc,
        "storage_location": _post_store_loc["url"],
        "file_type": _shell_script_type,
        "authors": [_user],
    }

    return fdp_req.post_else_get(
        uri, "object", data=_object_data, params={"description": _desc}
    )


def store_namespace(
    uri: str,
    namespace_label: str,
    full_name: str = None,
    website: str = None
) -> str:
    """Create a namespace on the registry

    Parameters
    ----------
    uri : str
        endpoint of the registry
    namespace_label : str
        name of the Namespace
    full_name : str, optional
        full title of the namespace, by default None
    website : str, optional
        website relating to this namespace, by default None

    Returns
    -------
    str
        URL of the created namespace
    """
    _data = {
        "name": namespace_label,
        "full_name": full_name,
        "website": website
    }
    return fdp_req.post_else_get(
        uri, "namespace", data=_data, params={"name": namespace_label}
    )


def store_data_file(
    uri: str,
    repo_dir: str,
    data: typing.Dict,
    local_file: str,
    write_data_store: str,
    public: bool
) -> None:

    _root_store = get_write_storage(uri, write_data_store)

    _rel_path = os.path.relpath(local_file, write_data_store)

    if 'version' not in data['use']:
        raise fdp_exc.InternalError(
            f"Expected version number for '{local_file}' "
            "registry submission but none found"
        )

    _hash = calculate_file_hash(local_file)

    _storage_loc_data = {
        "path": _rel_path,
        "storage_root": _root_store,
        "public": public,
        "hash": _hash,
    }

    _search_data = {
        'hash': _hash
    }

    _post_store_loc = fdp_req.post_else_get(
        uri,
        "storage_location",
        data=_storage_loc_data,
        params=_search_data
    )

    _user = store_user(repo_dir, uri)

    if 'file_type' in data:
        _file_type = data['file_type']
    else:
        _file_type = os.path.splitext(local_file)[1]

    _file_type = create_file_type(
        uri, _file_type
    )

    # Namespace is read from the source information
    if 'namespace_name' not in data:
        raise fdp_exc.UserConfigError(
            f"Expected 'namespace_name' for item '{local_file}'"
        )

    _namespace_args = {
        "uri": uri,
        "namespace_label": data['namespace_name'],
        "full_name": data['namespace_full_name']
            if 'namespace_full_name' in data else None,
        "website": data['namespace_website']
            if 'namespace_website' in data else None
    }

    _namespace_url = store_namespace(**_namespace_args)

    _desc = data['description'] if 'description' in data else None

    _object_data = {
        "description": _desc,
        "file_type": _file_type,
        "storage_location": _post_store_loc,
        "authors": [_user]
    }

    try:
        _obj_url = fdp_req.post(uri, "object", data=_object_data)['url']
    except fdp_exc.RegistryAPICallError as e:
        if not e.error_code == 409:
            raise e
        else:
            raise fdp_exc.RegistryAPICallError(
                f"Cannot post object"
                f"'{_desc}', duplicate already exists",
                error_code=409
            )
    except KeyError:
        raise fdp_exc.InternalError(
            f"Expected key 'url' in local registry API response"
            f" for post object '{_desc}'"
        )

    # Get the name of the entry
    if 'external_object' in data:
        _name = data['external_object']
    elif 'data_product' in data:
        _name = data['data_product']
    else:
        raise fdp_exc.UserConfigError(
            f"Failed to determine type while storing item '{local_file}'"
            "into registry"
        )

    if 'data_product' in data['use']:
        _name = data['use']['data_product']

    _data_prod_data = {
        "namespace": _namespace_url,
        "object": _obj_url,
        "version": str(data['use']['version']),
        "name": _name
    }

    try:
        _data_prod_url = fdp_req.post(uri, 'data_product', data=_data_prod_data)['url']
    except fdp_exc.RegistryAPICallError as e:
        if not e.error_code == 409:
            raise e
        else:
            raise fdp_exc.RegistryAPICallError(
                f"Cannot post data_product "
                f"'{_name}', duplicate already exists",
                error_code=409
            )
    except KeyError:
        raise fdp_exc.InternalError(
            f"Expected key 'url' in local registry API response"
            f" for post object '{_name}'"
        )

    # If 'data_product' key present finish here and return URL
    # else this is an external object
    if 'data_product' in data:
        return _data_prod_url

    _expected_ext_obj_keys = [
        "release_date",
        "primary",
        "title"
    ]

    _identifier = None
    _alternate_identifier = None
    _alternate_identifier_type = None

    if 'identifier' in data:
        _identifier = data['identifier']
        if not fdp_id.check_id_permitted(_identifier):
            raise fdp_exc.UserConfigError(
                f"Identifier '{_identifier}' is not a valid identifier"
            )

    if not _identifier:
        if 'unique_name' not in data:
            raise fdp_exc.UserConfigError(
                "No identifier/alternate_identifier given for "
                f"item '{local_file}'",
                hint="You must provide either a URL 'identifier', or "
                "'unique_name' and 'source_name' keys"
            )
        else:
            _alternate_identifier = data['unique_name']
            if 'alternate_identifier_type' in data:
                _alternate_identifier_type = data['alternate_identifier_type']
            else:
                _alternate_identifier_type = 'local source descriptor'

    for key in _expected_ext_obj_keys:
        if key not in data:
            raise fdp_exc.UserConfigError(
                f"Expected key '{key}' for item '{local_file}'"
            )

    _external_obj_data = {
        "data_product": _data_prod_url,
        "title": data['title'],
        "primary_not_supplement": data['primary'],
        "release_date": data['release_date'],
        "identifier": _identifier,
        "alternate_identifier": _alternate_identifier,
        "alternate_identifier_type": _alternate_identifier_type
    }

    return fdp_req.post(uri, 'external_object', data=_external_obj_data)


def calculate_file_hash(file_name: str, buffer_size: int = 64*1024) -> str:
    """Calculates the hash of a data file

    Parameters
    ----------
    file_name : str
        file to calculate

    Returns
    -------
    str
        SHA1 hash for file
    """
    # If the file is large we do not want to hash it in one go
    _input_hasher = hashlib.sha1()

    with open(file_name, 'rb') as in_f:
        _buffer = in_f.read(buffer_size)
        while len(_buffer) > 0:
            _input_hasher.update(_buffer)
            _buffer = in_f.read(buffer_size)

    return _input_hasher.hexdigest()


def get_storage_root_obj_address(
    remote_uri: str,
    remote_token: str,
    address_str: str) -> str:
    """Retrieve the RestAPI URL for a given storage location on the registry

    Parameters
    ----------
    remote_uri : str
        endpoint of remote registry
    remote_token : str
        token for accessing remote registry
    address_str : str
        path of the storage location

    Returns
    -------
    str
        URL of the RestAPI object representing this address
    """
    try:
        _results = fdp_req.get(
            remote_uri,
            'storage_root',
            params={
                'root': address_str
            },
            token=remote_token
        )
        if not _results:
            raise AssertionError
    except (AssertionError, fdp_exc.RegistryAPICallError):
        raise fdp_exc.RegistryAPICallError(
            f"Cannot find a match for path '{address_str}' "
            f"from endpoint '{remote_uri}."
        )


def check_match(input_object: str, results_list: typing.List[str]):
    """Check if an object is already on the registry using a hash

    In order to perform this check the results must be a list of storage
    locations for a given list of data_products as SHA1 hashes are present
    for locations only.

    Parameters
    ----------
    input_object : str
        object pending submission to registry
    results_list : typing.List[str]
        list of storage_location results matching search query
    """
    _hash_list = [res['hash'] for res in results_list]
    return calculate_file_hash(input_object) in _hash_list


def check_if_object_exists(
    local_uri: str,
    file_loc: str,
    obj_type: str,
    search_data: typing.Dict,
    token: str = None) -> str:
    """Checks if a data product is already present in the registry

    Parameters
    ----------
    local_uri : str
        local registry endpoint
    file_loc : str
        path of file on system
    obj_type : str
        object type
    search_data : typing.Dict
        data for query
    token : str
        token for registry

    Returns
    -------
    str
        whether object is present with matching hash, present or absent
        if present but not hash matched return the latest version identifier
    """
    _version = None

    if 'version' in search_data:
        _version = search_data['version']
        if '${{' in _version:
            del search_data['version']

    # Obtain list of storage_locations for the given data_product
    _results = fdp_req.get(
        local_uri,
        obj_type,
        params=search_data,
        token=token
    )

    try:
        fdp_ver.get_correct_version(
            version=_version,
            results_list=_results,
            free_write=True
        )
    except fdp_exc.UserConfigError:
        return "absent"

    if not _results:
        return "absent"

    if obj_type == 'external_object':
        _results = [res['data_product'] for res in _results]
        _results = [fdp_req.url_get(r) for r in _results]

    _object_urls = [res['object'] for res in _results]

    _storage_urls = [
        fdp_req.url_get(obj_url, token=token)['storage_location']
        for obj_url in _object_urls
    ]

    _storage_objs = [
        fdp_req.url_get(store_url, token=token)
        for store_url in _storage_urls
    ]

    if check_match(file_loc, _storage_objs):
        return "hash_match"
    else:
        return _results



from collections.abc import MutableMapping
import typing
import yaml
import os.path
import logging
import copy
import os
import re
import git
import platform
import datetime

import fair.exceptions as fdp_exc
import fair.utilities as fdp_util
import fair.configuration as fdp_conf
import fair.registry.versioning as fdp_ver
import fair.parsing.globbing as fdp_glob
import fair.registry.storage as fdp_store
import fair.registry.requests as fdp_req
import fair.common as fdp_com
import fair.register as fdp_reg

from fair.common import CMD_MODE


JOB2CLI_MAPPINGS = {
    'run_metadata.local_repo': 'git.local_repo',
    'run_metadata.default_input_namespace': 'namespaces.input',
    'run_metadata.default_output_namespace': 'namespaces.output',
    'run_metadata.local_data_registry_url': 'registries.local.uri',
    'run_metadata.remote_data_registry_url': 'registries.origin.uri',
    'run_metadata.write_data_store': 'registries.local.data_store',
}


class JobConfiguration(MutableMapping):
    _logger = logging.getLogger('FAIRDataPipeline.ConfigYAML')
    _block_types = ('register', 'write', 'read')
    def __init__(self, config_yaml: str) -> None:
        if not os.path.exists(config_yaml):
            raise fdp_exc.FileNotFoundError(
                f"Cannot load job configuration from file '{config_yaml}', "
                "file does not exist"
            )

        self._logger.debug("Loading file '%s'", config_yaml)

        self._config: typing.Dict = yaml.safe_load(open(config_yaml))

        self._fill_missing()

        self.env = None

    def __contains__(self, key_addr: str) -> bool:
        return key_addr in fdp_util.flatten_dict(self._config)

    def __setitem__(self, key_addr: str, value: typing.Any, separator: str = '.') -> None:
        if key_addr in self._config:
            self._config[key_addr] = value
        _flat_cfg = fdp_util.flatten_dict(self._config, separator)
        _flat_cfg[key_addr] = value
        self._logger.debug(f"Set value {key_addr.replace('.', ':')}={value}'")
        self._config = fdp_util.expand_dict(_flat_cfg)

    def __delitem__(self, key_addr: str, separator: str = '.') -> None:
        if key_addr in self._config:
            del self._config[key_addr]
        _flat_cfg = fdp_util.flatten_dict(self._config, separator)
        if key_addr not in _flat_cfg:
            raise fdp_exc.KeyPathError(key_addr, fdp_com.USER_CONFIG_FILE)
        self._logger.debug(f"Removing '{key_addr}'")
        del _flat_cfg[key_addr]
        self._config = fdp_util.expand_dict(_flat_cfg)
    
    def __getitem__(self, key_addr: str, separator: str = '.') -> None:
        if key_addr in self._config:
            return self._config[key_addr]
        _flat_cfg = fdp_util.flatten_dict(self._config, separator)
        if key_addr not in _flat_cfg:
            raise fdp_exc.KeyPathError(key_addr, fdp_com.USER_CONFIG_FILE)
        return _flat_cfg[key_addr]
    
    def __len__(self) -> int:
        raise fdp_exc.NotImplementedError(
            "Cannot get length of a Job Configuration, attribute has no meaning in this context"
        )

    def __iter__(self) -> typing.Any:
        raise fdp_exc.NotImplementedError(
            "Cannot iterate through a Job Configuration, operation is ambiguous"
        )

    def _fill_missing(self) -> None:
        self._logger.debug("Filling missing metadata")
        if 'run_metadata' not in self._config:
            self._logger.debug("Failed to find 'run_metadata' in configuration, creating")
            self['run_metadata'] = {}

        _required = [
            ('run_metadata.public', True),
            ('run_metadata.default_read_version', '${{'+fdp_ver.DEFAULT_READ_VERSION+'}}'),
            ('run_metadata.default_write_version', '${{'+fdp_ver.DEFAULT_WRITE_VERSION+'}}'),
        ]

        for item in _required:
            try:
                self[item[0]]
            except fdp_exc.KeyPathError:
                self[item[0]] = item[1]

    def _handle_register_namespaces(self) -> typing.Dict:
        self._logger.debug("Handling 'register' namespaces")
        _new_register_block: typing.List[typing.Dict] = []
        for item in self['register']:
            if any(i in item for i in ('external_object', 'data_product')):
                if 'namespace_name' in item:
                    _namespace_store_args = {
                        'namespace_label': item['namespace_name'],
                        'full_name': item.get('namespace_full_name', None),
                        'website': item.get('namespace_website', None)
                    }
                elif 'namespace' in item and isinstance(item['namespace'], dict):
                    _namespace_store_args = item['namespace']
                _new_register_block.append(item)
            elif 'namespace' in item:
                _namespace_store_args = {
                    'namespace_label': item['namespace'],
                    'full_name': item.get('full_name', None),
                    'website': item.get('website', None)
                }
            else:
                _new_register_block.append(item)
            fdp_store.store_namespace(
                self.local_uri,
                **_namespace_store_args
            )
        return _new_register_block

    def _unpack_register_namespaces(self) -> None:
        self._logger.debug("Unpacking 'register' namespaces")
        for i, item in enumerate(self._config['register']):
            if all(it not in item for it in ['external_object', 'data_product']):
                continue
            
            if 'namespace' not in item:
                continue

            if isinstance(item['namespace'], str):
                self._config['register'][i]['namespace_name'] = item['namespace']
            elif isinstance(item['namespace'], dict):
                self._config['register'][i]['namespace_name'] = item['namespace']['name']
                self._config['register'][i]['namespace_full_name'] = None
                self._config['register'][i]['namespace_website'] = None
                del self._config['register'][i]['namespace']

    def _fill_namespaces(self, block_type: str) -> typing.List[typing.Dict]:
        self._logger.debug("Filling all namespaces")
        _entries: typing.List[typing.Dict] = []
        
        if block_type == 'read':
            _default_ns = self.default_input_namespace
        else:
            _default_ns = self.default_output_namespace

        for item in self[block_type]:
            if all(i not in item for i in ('data_product', 'external_object')):
                _entries.append(item)
                continue
            _new_item = copy.deepcopy(item)
            _new_item['use'] = item.get('use', {})

            # --------------------------------------------------------------- #
            # FIXME: Schema needs to be finalised, allowing item:namespace and
            # also item:use:namespace will cause confusion.

            if 'namespace' in item and 'namespace' not in _new_item['use']:
                _new_item['use']['namespace'] = _new_item['namespace']
                _new_item.pop('namespace')

            # --------------------------------------------------------------- #

            if 'namespace' not in _new_item['use']:
                _new_item['use']['namespace'] = _default_ns
            
            _entries.append(_new_item)
        
        return _entries


    def _update_namespaces(self) -> None:
        self._logger.debug("Updating namespace list")
        if 'register' in self._config:
            self['register'] = self._handle_register_namespaces()

        if not self.default_input_namespace:
            raise fdp_exc.UserConfigError(
                "Input namespace cannot be None"
            )

        if not self.default_output_namespace:
            raise fdp_exc.UserConfigError(
                "Output namespace cannot be None"
            )

        fdp_store.store_namespace(self.local_uri, self.default_input_namespace)
        fdp_store.store_namespace(self.local_uri, self.default_output_namespace)

        for block_type in self._block_types:
            if block_type not in self:
                continue
            self[block_type] = self._fill_namespaces(block_type)

        if 'register' in self._config:
            self._unpack_register_namespaces()

    def _expand_wildcards_from_local_reg(self, block_type: str) -> None:
        self._logger.debug("Expanding wildcards using local registry")
        _version = self.default_read_version if block_type == 'read' else self.default_write_version
        fdp_glob.glob_read_write(
            user_config=self._config,
            blocktype=block_type,
            search_key=None,
            registry_url = self.local_uri,
            version=_version,
            remove_wildcard = block_type == 'read'
        )

    def _fetch_latest_commit(self, allow_dirty: bool = False) -> None:
        self._logger.debug(f"Retrieving latest commit SHA with allow_dirty={allow_dirty}")
        _repository = git.Repo(fdp_com.find_git_root(self.local_repository))
        
        try:
            _latest = _repository.head.commit.hexsha
        except git.InvalidGitRepositoryError:
            raise fdp_exc.FDPRepositoryError(
                f"Location '{self.local_repository}' is not a valid git repository"
            )
        except ValueError:
            raise fdp_exc.FDPRepositoryError(
                f"Failed to retrieve latest commit for local repository '{self.local_repository}'",
                hint="Have any changes been committed in the project repository?"
            )
        
        if _repository.is_dirty():
            if not allow_dirty:
                raise fdp_exc.FDPRepositoryError(
                    "Repository contains uncommitted changes"
                )
            _latest = f'{_latest}-dirty'

        return _latest

    def update_from_fair(self, fair_repo_dir: str = None, remote_label: str = None) -> None:
        """Update any missing entries from defaults defined by FAIR CLI configurations
        
        Reads CLI configurations both globally and optionally from a specified FAIR
        repository and updates assigning any keys

        Parameters
        ----------
        fair_repo_dir : str, optional
            local FAIR repository directory
        remote_label : str, optional
            specify alternate remote registry by label, default is 'origin'
        """
        self._logger.debug(f"Updating configuration from {fair_repo_dir}")
        if fair_repo_dir and not os.path.exists(fair_repo_dir):
            raise fdp_exc.FileNotFoundError(
                "Cannot update configuration from repository, location"
                f" '{fair_repo_dir}' does not exist"
            )

        if fair_repo_dir:
            # Combine global and remote CLI configurations by overwriting global
            # with changes in the local version
            _fdpconfig = fdp_util.flatten_dict(fdp_conf.read_global_fdpconfig())
            _local_fdpconfig = fdp_util.flatten_dict(
                fdp_conf.read_local_fdpconfig(fair_repo_dir)
            )
            _fdpconfig.update(_local_fdpconfig)
        else:
            _fdpconfig = fdp_util.flatten_dict(fdp_conf.read_global_fdpconfig())

        for key in JOB2CLI_MAPPINGS:
            if key not in self:
                self[key] = _fdpconfig[JOB2CLI_MAPPINGS[key]]

        if remote_label:
            _key = 'run_metadata.remote_data_registry_url'
            _other_key = JOB2CLI_MAPPINGS[_key].replace('origin', remote_label)
            self[_key] = _fdpconfig[_other_key]

        if fair_repo_dir and 'run_metadata.remote_repo' not in self:
            _remote = _fdpconfig['git.remote']
            _git_repo = git.Repo(fair_repo_dir)
            _url = _git_repo.remotes[_remote].url
            self['run_metadata.remote_repo'] = _url

    def pop(self, key: str, default: typing.Optional[typing.Any] = None) -> typing.Any:
        """Remove item from configuration"""
        try:
            del self[key]
        except fdp_exc.KeyPathError:
            return default

    def get(self, key: str, default: typing.Optional[typing.Any] = None) -> typing.Any:
        """Retrieve item if exists, else return default"""
        try:
            _value = self[key]
        except fdp_exc.KeyPathError:
            _value = default
        self._logger.debug(f"Returning '{key}={_value}'")
        return _value

    def set_command(self, cmd: str, shell: typing.Optional[str] = None) -> None:
        """Set a BASH command to be executed"""
        if not shell:
            shell = "batch" if platform.system() == "Windows" else "sh"
        self._logger.debug(f"Setting {shell} command to '{cmd}'")
        self['run_metadata.script'] = cmd
        self['run_metadata.shell'] = shell
        self.pop('run_metadata.script_path')

    def _create_environment(self, output_dir: str) -> None:
        """Create the environment for running a job"""
        _environment = os.environ.copy()
        _environment["FDP_LOCAL_REPO"] = self.local_repository
        _environment["FDP_CONFIG_DIR"] = output_dir
        _environment["FDP_LOCAL_TOKEN"] = fdp_req.local_token()
        return _environment

    def set_script(self, command_script: str) -> None:
        """Set a BASH command to be executed"""
        self._logger.debug(f"Setting command script to '{command_script}'")
        self['run_metadata.script_path'] = command_script
        self.pop('run_metadata.script')

    def _substitute_variables(self, job_dir: str, time_stamp: datetime.datetime) -> None:
        self._logger.debug("Performing variable substitution")
        _config_str = self._subst_cli_vars(job_dir, time_stamp)
        self._config = yaml.safe_load(_config_str)

    def prepare(self, job_dir: str, time_stamp: datetime.datetime, job_mode: CMD_MODE) -> None:
        """Initiate a job execution"""
        self._logger.debug("Preparing configuration")
        self._update_namespaces()
        self._substitute_variables(job_dir, time_stamp)
        
        self._fill_all_block_types()

        if job_mode in [CMD_MODE.PULL, CMD_MODE.PUSH]:
            self._pull_metadata()

        for block_type in ('read', 'write'):
            if block_type not in self:
                continue
            try:
                self._expand_wildcards_from_local_reg(block_type)
            except fdp_exc.InternalError:
                continue

        for block_type in self._block_types:
            if block_type in self:
                self[block_type] = self._fill_versions(block_type)

        if 'register' in self:
            if 'read' not in self:
                self['read'] = []
            self['read'] += self._register_to_read()
        
        if job_mode in [CMD_MODE.PULL, CMD_MODE.PUSH]:
            self._pull_data()

        if job_mode == CMD_MODE.PULL and 'register' in self:
            _objs = fdp_reg.fetch_registrations(
                local_uri=self.local_uri,
                repo_dir=self.local_repository,
                write_data_store=self.default_data_store,
                user_config_register=self['register']
            )
            self._logger.debug("Fetched objects:\n %s", _objs)
        
        self._config = self._clean()

        _unparsed =  self._check_for_unparsed()

        if _unparsed:
            raise fdp_exc.InternalError(
                f"Failed to parse variables '{_unparsed}'"
            )

        self['run_metadata.latest_commit'] = self._fetch_latest_commit()

    def _check_for_unparsed(self) -> typing.List[str]:
        self._logger.debug("Checking for unparsed variables")
        _conf_str = yaml.dump(self._config)

        # Additional parser for formatted datetime
        _regex_fmt = re.compile(r'\$\{\{\s*([^}${\s]+)\s*\}\}')

        return _regex_fmt.findall(_conf_str)

    def _subst_cli_vars(self, job_dir: str, job_time: datetime.datetime) -> str:
        self._logger.debug("Searching for CLI variables")

        def _get_id():
            try:
                return fdp_conf.get_current_user_uri(job_dir)
            except fdp_exc.CLIConfigurationError:
                return fdp_conf.get_current_user_uuid(job_dir)

        def _tag_check(*args, **kwargs):
            _repo = git.Repo(fdp_conf.local_git_repo(self.local_repository))
            if len(_repo.tags) < 1:
                fdp_exc.UserConfigError(
                    "Cannot use GIT_TAG variable, no git tags found."
                )
            return _repo.tags[-1].name

        _substitutes: typing.Dict[str, typing.Callable] = {
            "DATE": lambda : job_time.strftime("%Y%m%d"),
            "DATETIME": lambda : job_time.strftime("%Y-%m-%dT%H:%M:%S%Z"),
            "USER": lambda : fdp_conf.get_current_user_name(self.local_repository),
            "USER_ID": lambda : _get_id(),
            "REPO_DIR": lambda : self.local_repository,
            "CONFIG_DIR": lambda : job_dir + os.path.sep,
            "LOCAL_TOKEN": lambda : fdp_req.local_token(),
            "GIT_BRANCH": lambda : self.git_branch,
            "GIT_REMOTE": lambda : self.git_remote_uri,
            "GIT_TAG": _tag_check,
        }

        # Additional parser for formatted datetime
        _regex_dt_fmt = re.compile(r'\$\{\{\s*DATETIME\-[^}${\s]]+\s*\}\}')
        _regex_fmt = re.compile(r'\$\{\{\s*DATETIME\-([^}${\s]+)\s*\}\}')

        _config_str: str = yaml.dump(self._config)

        _dt_fmt_res: typing.Optional[typing.List[str]] = _regex_dt_fmt.findall(_config_str)
        _fmt_res: typing.Optional[typing.List[str]] = _regex_fmt.findall(_config_str)

        self._logger.debug(
            "Found datetime substitutions: %s %s",
            _dt_fmt_res or "",
            _fmt_res or ""
        )

        # The two regex searches should match lengths
        if len(_dt_fmt_res) != len(_fmt_res):
            raise fdp_exc.UserConfigError(
                "Failed to parse formatted datetime variable"
            )

        if _dt_fmt_res:
            for i, _ in enumerate(_dt_fmt_res):
                _time_str = job_time.strftime(_fmt_res[i].strip())
                _config_str = _config_str.replace(_dt_fmt_res[i], _time_str)

        _regex_dict = {
            var: r'\$\{\{\s*'+f'{var}'+r'\s*\}\}'
            for var in _substitutes
        }

        # Perform string substitutions
        for var, subst in _regex_dict.items():
            # Only execute functions in var substitutions that are required
            if re.findall(subst, _config_str):
                _value = _substitutes[var]()
                if not _value:
                    raise fdp_exc.InternalError(
                        f"Expected value for substitution of '{var}' but returned None",
                    )
                _config_str = re.sub(subst, str(_value), _config_str)
                self._logger.debug("Substituting %s: %s", var, str(_value))
        
        # Load the YAML (this also verifies the write was successful) and return it
        return _config_str

    def _register_to_read(self) -> typing.Dict:
        """Construct 'read' block entries from 'register' block entries
        
        Parameters
        ----------
        register_block : typing.Dict
            register type entries within a config.yaml
        
        Returns
        -------
        typing.Dict
            new read entries extract from register statements
        """
        _read_block: typing.List[typing.Dict] = []

        for item in self._config['register']:
            _readable = {}
            if 'use' in item:
                _readable['use'] = copy.deepcopy(item['use'])
            if 'external_object' in item:
                _readable['data_product'] = item['external_object']
            elif 'data_product' in item:
                _readable['data_product'] = item['data_product']
            elif 'namespace' in item:
                fdp_store.store_namespace(**item)
            else: # unknown
                raise fdp_exc.UserConfigError(
                    f"Found registration for unknown item with keys {[*item]}"
                )
            _readable['use']['version'] = fdp_ver.undo_incrementer(_readable['use']['version'])
            
            _read_block.append(_readable)

        return _read_block

    def _clean(self) -> typing.Dict:
        self._logger.debug("Cleaning configuration")
        _new_config: typing.Dict = {}
        _new_config['run_metadata'] = copy.deepcopy(self['run_metadata'])

        for action in ('read', 'write'):
            if f'default_{action}_version' in _new_config['run_metadata']:
                del _new_config['run_metadata'][f'default_{action}_version']

        _namespaces = (
            self.default_input_namespace,
            self.default_output_namespace
        )

        for namespace, block_type in zip(_namespaces, ['read', 'write']):
            if block_type not in self._config:
                continue

            _new_config[block_type] = []

            for item in self[block_type]:
                for use_item in [*item['use']]:
                    # Get rid of duplicates
                    if use_item in item and item['use'][use_item] == item[use_item]:
                        item['use'].pop(use_item)
                
                if block_type == 'write' and item['public'] == self.is_public_global:
                    item.pop('public')

                if item['use']['namespace'] == namespace:
                    item['use'].pop('namespace')

                _new_config[block_type].append(item)        

        for block in self._block_types:
            if block in _new_config and not _new_config[block]:
                del _new_config[block]

        return _new_config

    def _fill_versions(self, block_type: str) -> typing.List[typing.Dict]:
        self._logger.debug("Filling version information")
        _entries: typing.List[typing.Dict] = []

        if block_type == 'read':
            _default_ver = self.default_read_version
        else:
            _default_ver = self.default_write_version

        for item in self._config[block_type]:
            if all(i not in item for i in ('data_product', 'external_object')):
                _entries.append(item)
                continue
            _new_item = copy.deepcopy(item)
            _new_item['use'] = item.get('use', {})

            # --------------------------------------------------------------- #
            # FIXME: Schema needs to be finalised, allowing item:version and
            # also item:use:version will cause confusion.

            if 'version' in item and 'version' not in _new_item['use']:
                _new_item['use']['version'] = _new_item['version']
                _new_item.pop('version')

            # --------------------------------------------------------------- #

            if 'version' not in _new_item['use']:
                _new_item['use']['version'] = _default_ver
            
            _version = item['use']['version']

            if 'data_product' not in item['use']:
                if 'external_object' in item and '*' in item['external_object']:
                    _name = item['external_object']
                elif 'data_product' in item and '*' in item['data_product']:
                    _name = item['data_product']
                else:
                    self._logger.warning(f'Missing use:data_product in {item}')
                continue

            _name = item['use']['data_product']
            _namespace = item['use']['namespace']

            if '${{' in _version:
                _results = fdp_req.get(self.local_uri, 'data_product', params = {
                    'name': _name, 'namespace': _namespace
                })
            else:
                _results = fdp_req.get(self.local_uri, 'data_product', params = {
                    'name': _name, 'namespace': _namespace, 'version': _version
                })

            try:
                _version = fdp_ver.get_correct_version(
                    version=_version,
                    results_list=_results,
                    free_write=block_type != 'read' 
                )
            except fdp_exc.UserConfigError as e:
                if block_type == 'register':
                    self._logger.debug(
                        f"Already found this version ({e}), but may be identical"
                    )
                else:
                    self._logger.error(str(item))
                    raise e

            if '${{' in _version:
                self._logger.error(
                    f"Found a version ({_version}) that needs resolving"
                )

            if str(_version) != item['use']['version']:
                item['use']['version'] = str(_version)
            
            _entries.append(item)
        
        return _entries


    def _pull_metadata(self) -> None:
        self._logger.debug("Pulling metadata from remote registry")
        self._logger.warning(
            "Remote registry pulls are not yet implemented"
        )

    def _pull_data(self) -> None:
        self._logger.debug("Pulling data from remote registry")
        self._logger.warning(
            "Remote registry pulls are not yet implemented"
        )

    def _fill_block_data_product(self, block_type: str, item: typing.Dict) -> typing.Dict:
        _new_item = copy.deepcopy(item)
        if 'namespace' in item:
            _new_item['use']['namespace'] = item['namespace']
            _new_item.pop('namespace')

        if 'namespace' not in item['use']:
            if block_type == "read":
                item['use']['namespace'] = self.default_input_namespace
            elif block_type == "write":
                item['use']['namespace'] = self.default_output_namespace
            elif block_type == "register":
                item['use']['namespace'] = item['namespace_name']
            else:
                raise fdp_exc.NotImplementedError(
                    f"Failed to understand block type '{block_type}'"
                )

        if 'version' not in item['use']:
            if 'version' in item:
                _new_item['use']['version'] = item['version']
                _new_item.pop('version')
            elif block_type == 'read':
                _new_item['use']['version'] = self.default_read_version
            else: # 'write' or 'register'
                _new_item['use']['version'] = self.default_write_version

        if 'data_product' not in item['use'] and '*' not in item['data_product']:
            _new_item['use']['data_product'] = item['data_product']

        if block_type == 'register' and 'external_object' in item:
            _new_item.pop('data_product', None)

        return _new_item

    def _fill_block_item(self, block_type: str, item: typing.Dict) -> typing.Dict:
        _new_item = copy.deepcopy(item)
        _new_item['use'] = item.get('use', {})

        _reg_obj = [
            block_type == 'register',
            'external_object' in item,
            'data_product' not in item,
            'data_product' not in _new_item['use']
        ]

        if all(_reg_obj):
            _new_item['use']['data_product'] = item['external_object']

        if 'data_product' in item:
            _new_item = self._fill_block_data_product(block_type, _new_item)

        if block_type in {'write', 'register'} and 'public' not in item:
            _new_item['public'] = self.is_public_global

        return _new_item

    def _fill_all_block_types(self) -> bool:
        for block_type in self._block_types:
            self._logger.debug(f"Filling '{block_type}' block")
            _new_block: typing.List[typing.Dict] = []
            if block_type not in self:
                continue
        
            for item in self[block_type]:
                _new_item = self._fill_block_item(block_type, item)
                _new_block.append(_new_item)

            self._config[block_type] = _new_block

    @property
    def content(self) -> typing.Dict:
        """Return a copy of the internal dictionary"""
        return copy.deepcopy(self._config)

    @property
    def shell(self) -> str:
        """Retrieve the shell choice"""
        _shell_default = "batch" if platform.system() == "Windows" else "sh"
        return self.get('shell', _shell_default)

    @property
    def local_repository(self) -> str:
        """Retrieves the local project repository"""
        return self['run_metadata.local_repo']

    @property
    def default_data_store(self) -> str:
        """Retrieves the default write data store"""
        return self['run_metadata.write_data_store']

    @property
    def default_input_namespace(self) -> str:
        """Retrieves the default write data store"""
        return self['run_metadata.default_input_namespace']

    @property
    def default_output_namespace(self) -> str:
        """Retrieves the default write data store"""
        return self['run_metadata.default_output_namespace']

    @property
    def default_read_version(self) -> str:
        """Retrieves the default read version"""
        return self.get('run_metadata.default_read_version', '${{'+fdp_ver.DEFAULT_READ_VERSION+'}}')

    @property
    def default_write_version(self) -> str:
        """Retrieves the default write version"""
        return self.get('run_metadata.default_write_version', '${{'+fdp_ver.DEFAULT_WRITE_VERSION+'}}')

    @property
    def is_public_global(self) -> bool:
        """Retrieves global publicity setting"""
        return self.get('run_metadata.public', True)

    @property
    def local_uri(self) -> str:
        """Retrieves the local URI for registry for this session"""
        return self.get('run_metadata.local_data_registry_url', fdp_conf.get_local_uri())

    @property
    def remote_uri(self) -> str:
        """Retrieves the remote URI for registry for this session"""
        return self.get(
            'run_metadata.remote_data_registry_url',
            fdp_conf.get_remote_uri(self.local_repository)
        )

    @property
    def git_remote_uri(self) -> str:
        """Retrieves the URI of the remote repository on git"""
        return self['run_metadata.remote_repo']

    @property
    def git_branch(self) -> str:
        """Retrieves the current git repository branch"""
        _git_loc = fdp_conf.local_git_repo(self.local_repository)
        return git.Repo(_git_loc).active_branch.name

    @property
    def command(self) -> typing.Optional[str]:
        """Returns either the script or script path to be executed"""
        for key in ('script', 'script_path'):
            if key in self['run_metadata']:
                return self[f'run_metadata.{key}']
        return None

    @property
    def environment(self) -> typing.Dict:
        """Returns the job execution environment"""
        return self.env

    def write(self, output_file: str) -> None:
        """Write job configuration to file"""
        with open(output_file, 'w') as out_f:
            yaml.dump(self._config, out_f)
        
        self.env = self._create_environment(os.path.dirname(output_file))
        
        self._logger.debug(f"Configuration written to '{output_file}'")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
"""

Configuration
=============

Module contains methods relating to configuration of the CLI itself.

Contents
========

Functions
---------

    read_local_fdpconfig - read the contents of the local CLI config file
    read_global_fdpconfig - read the contents of the global CLI config file
    set_email - set the user's email in the configuration
    set_user - set the user's name in the configuration
    get_current_user_name - retrieve name of the current user
    get_current_user_email - retrieve email of the current user
    get_remote_uri - retrieve URL for the remote registry
    local_git_repo - retrieve path for the current project Git repository
    remote_git_repo - retrieve URL for the remote Git repository
    get_remote_token - retrieve the token for the remote registry
    get_session_git_remote - get the current git remote label or URL
    get_current_user_uri - retrieve the full URL identifier for the current user
    get_current_user_uuid - retrieve the uuid of the current user if specified
    check_registry_exists - check that the specified local registry directory exists
    get_local_uri - retrieve the URL of the local registry
    get_local_port - retrieve the port number of the local registry
    global_config_query - setup the global configuration using prompts
    local_config_query - setup the local configuration using prompts
    write_data_store - retrieve the current write data store
    input_namespace - retrieve current input namespace
    output_namespace - retrieve current output namespace
    is_public - retrieve the current data is_public label
    read_version - retrieve the current data read version
    write_version - retrieve the current data write
    registry_url - retrieves either local or remote registry URL
    update_metadata - updates the metadata within a use configuration file
"""

__date__ = "2021-07-02"

import os
import platform
import pathlib
import uuid
import copy
import logging
import typing
import yaml
import click
import git

from urllib.parse import urlparse, urljoin

import fair.common as fdp_com
import fair.exceptions as fdp_exc
import fair.identifiers as fdp_id
import fair.registry.server as fdp_serv
import fair.registry.requests as fdp_req
import fair.registry.versioning as fdp_ver

logger = logging.getLogger('FAIRDataPipeline.Configuration')


def read_local_fdpconfig(repo_loc: str) -> typing.MutableMapping:
    """Read contents of repository level FAIR-CLI configurations.

    Parameters
    ----------
    repo_loc : str
        FAIR-CLI repository directory

    Returns
    -------
    MutableMapping
        configurations as a mapping
    """
    _local_config: typing.MutableMapping = {}

    # Retrieve the location of this repositories CLI config file
    _local_config_file_addr = fdp_com.local_fdpconfig(repo_loc)
    if os.path.exists(_local_config_file_addr):
        _local_config = yaml.safe_load(open(_local_config_file_addr))

    return _local_config


def read_global_fdpconfig() -> typing.MutableMapping:
    """Read contents of the global FAIR-CLI configurations.

    Returns
    -------
    MutableMapping
        configurations as a mapping
    """
    _global_config: typing.MutableMapping = {}

    # Retrieve the location of the global CLI config file
    _global_config_addr = fdp_com.global_fdpconfig()

    if os.path.exists(_global_config_addr):
        _global_config = yaml.safe_load(open(_global_config_addr))

    return _global_config


def set_email(repo_loc: str, email: str, is_global: bool = False) -> None:
    """Update the email address for the user

    Parameters
    ----------
    repo_loc : str
        repository directory path
    email : str
        new email address to set
    is_global : bool, optional
        whether to also override the global settings, by default False
    """
    _loc_conf = read_local_fdpconfig(repo_loc)
    _loc_conf['user']['email'] = email
    yaml.dump(_loc_conf, open(fdp_com.local_fdpconfig(repo_loc), "w"))
    if is_global:
        _glob_conf = read_global_fdpconfig()
        _glob_conf['user']['email'] = email
        yaml.dump(_glob_conf, open(fdp_com.global_fdpconfig(), "w"))


def set_user(repo_loc: str, name: str, is_global: bool = False) -> None:
    """Update the name for the user

    Parameters
    ----------
    repo_loc : str
        repository directory path
    name : str
        new user full name
    is_global : bool, optional
        whether to also override the global settings, by default False
    """
    _loc_conf = read_local_fdpconfig(repo_loc)
    if len(name.split()) > 1:
        _given_name, _family_name = name.rsplit(" ", 1)
        _loc_conf['user']['given_names'] = _given_name.title().strip()
        _loc_conf['user']['family_name'] = _family_name.title().strip()
        if is_global:
            _glob_conf = read_global_fdpconfig()
            _glob_conf['user']['given_names'] = _given_name.title().strip()
            _glob_conf['user']['family_name'] = _family_name.title().strip()
            yaml.dump(_glob_conf, open(fdp_com.global_fdpconfig(), "w"))
    else:
        _loc_conf['user']['given_names'] = name.title().strip()
        _loc_conf['user']['family_name'] = None
        if is_global:
            _glob_conf = read_global_fdpconfig()
            _glob_conf['user']['given_names'] = name.title().strip()
            _glob_conf['user']['family_name'] = None
            yaml.dump(_glob_conf, open(fdp_com.global_fdpconfig(), "w"))
    yaml.dump(_loc_conf, open(fdp_com.local_fdpconfig(repo_loc), "w"))
    if is_global:
        _glob_conf = read_global_fdpconfig()
        _glob_conf['user']['name'] = name
        yaml.dump(_glob_conf, open(fdp_com.global_fdpconfig(), "w"))         


def get_current_user_name(repo_loc: str) -> typing.Tuple[str]:
    """Retrieves the name of the current session user as defined in the config

    Returns
    -------
    str
        user name
    """
    _local_conf = read_local_fdpconfig(repo_loc)

    if not _local_conf:
        raise fdp_exc.CLIConfigurationError("Cannot retrieve current user from empty CLI config")

    _given = _local_conf['user']['given_names']
    if "family_name" in _local_conf['user']:
        _family = _local_conf['user']['family_name']
    else:
        _family = ""
    return (_given, _family)


def get_current_user_email(repo_loc: str) -> typing.Tuple[str]:
    """Retrieves the email of the current session user as defined in the config

    Returns
    -------
    str
        user email
    """
    _local_conf = read_local_fdpconfig(repo_loc)

    if not _local_conf:
        raise fdp_exc.CLIConfigurationError("Cannot retrieve current user from empty CLI config")

    return _local_conf['user']['email']


def get_remote_uri(repo_loc: str, remote_label: str = 'origin') -> str:
    """Retrieves the URI of the remote registry

    Parameters
    ----------
    repo_loc : str
        local FAIR repository directory
    remote_label : str, optional
        label of remote to retrieve, default is 'origin'

    Returns
    -------
    str
        remote URI path
    """
    _local_conf = read_local_fdpconfig(repo_loc)
    return _local_conf['registries'][remote_label]['uri']


def local_git_repo(fair_repo_loc: str) -> str:
    """Retrieves the local repository git directory

    Parameters
    ----------
    fair_repo_loc : str
        FAIR repository location

    Returns
    -------
    str
        the root git repository directory
    """
    _local_conf = read_local_fdpconfig(fair_repo_loc)

    try:
        return _local_conf['git']['local_repo']
    except KeyError:
        raise fdp_exc.CLIConfigurationError(
            "Expected key 'git:local_repo' in local CLI configuration"
        )


def remote_git_repo(fair_repo_loc: str) -> str:
    """Retrieves the remote repository git directory

    Parameters
    ----------
    fair_repo_loc : str
        FAIR repository location

    Returns
    -------
    str
        the git repository remote URL
    """

    _local_conf = read_local_fdpconfig(fair_repo_loc)

    try:
        return _local_conf['git']['remote_repo']
    except KeyError:
        raise fdp_exc.CLIConfigurationError(
            "Expected key 'git:remote_repo' in local CLI configuration"
        )


def get_remote_token(repo_dir: str, remote: str = 'origin') -> str:
    _local_config = read_local_fdpconfig(repo_dir)
    if remote not in _local_config['registries']:
        raise fdp_exc.CLIConfigurationError(
            f"Cannot find remote registry '{remote}' in local CLI configuration"
        )
    if 'token' not in _local_config['registries'][remote]:
        raise fdp_exc.CLIConfigurationError(
            f"Cannot find token for registry '{remote}', no token file provided"
        )
    _token_file = _local_config['registries'][remote]['token']

    if not os.path.exists(_token_file):
        raise fdp_exc.FileNotFoundError(
            f"Cannot read token for registry '{remote}', no such token file"
        )

    _token = open(_token_file).read().strip()

    if not _token:
        raise fdp_exc.CLIConfigurationError(
            f"Cannot read token from file '{_token_file}', file is empty."
        )

    return _token


def get_local_data_store() -> str:
    """Retrieves the local data store path"""
    _global_conf = read_global_fdpconfig()

    try:
        return _global_conf['registries']['local']['data_store']
    except KeyError:
        raise fdp_exc.CLIConfigurationError(
            "Expected key 'registries:local:data_store' in global CLI configuration"
        )


def get_session_git_remote(repo_loc: str, url: bool = False) -> str:
    """Retrieves the local repository git remote

    Parameters
    ----------
    repo_loc : str
        Location of session CLI config
    url : optional, bool
        If True return the URL for the remote instead

    Returns
    -------
    str
        the remote label or URL of the git repository
    """
    _local_conf = read_local_fdpconfig(repo_loc)

    try:
        _remote_label = _local_conf['git']['remote']
    except KeyError:
        raise fdp_exc.InternalError(
            "Failed to retrieve project git repository remote"
        )
    
    if not url:
        return _remote_label
    
    _repo_root = fdp_com.find_git_root(repo_loc)

    try:
        return git.Repo(_repo_root).remote(_remote_label).url
    except ValueError:
        raise fdp_exc.CLIConfigurationError(
            f"Failed to retrieve URL for git remote '{_remote_label}'"
        )


def get_current_user_uuid(repo_loc: str) -> str:
    """Retrieves the UUID of the current session user if defined in the config

    Returns
    -------
    str
        user UUID
    """
    _local_conf = read_local_fdpconfig(repo_loc)
    try:
        _uuid = _local_conf['user']['uuid']
    except KeyError:
        _uuid = None
    if not _uuid or _uuid == "None":
        raise fdp_exc.CLIConfigurationError("No UUID defined.")
    return _uuid

def get_current_user_uri(repo_loc: str) -> str:
    """Retrieves the URI identifier for the current user

    Returns
    -------
    str
        user URI identifier
    """
    _local_conf = read_local_fdpconfig(repo_loc)
    try:
        _uri = _local_conf['user']['uri']
    except KeyError:
        _uri = None
    if not _uri or _uri == "None":
        raise fdp_exc.CLIConfigurationError("No user URI identifier defined.")
    return _uri


def check_registry_exists(registry: str = None) -> bool:
    """Checks if fair registry is set up on users machine

    Returns
    -------
    bool
        True if registry exists, else False
    """
    if not registry:
        registry = fdp_serv.DEFAULT_REGISTRY_LOCATION
    return os.path.isdir(registry)


def get_local_uri() -> str:
    """Retrieve the local registry URI"""
    _global_conf = read_global_fdpconfig()

    try:
        return _global_conf['registries']['local']['uri']
    except KeyError:
        raise fdp_exc.CLIConfigurationError(
            "Expected key 'registries:local:uri' in global CLI configuration"
        )


def get_local_port(local_uri: str = None) -> int:
    """Retrieve the local port from the local URI"""
    if not local_uri:
        local_uri = get_local_uri()
    _port = urlparse(local_uri).port
    if not _port:
        raise fdp_exc.InternalError(
            "Failed to determine port number from local registry URL"
        )
    return _port

def _handle_orcid(user_orcid: str) -> typing.Tuple[typing.Dict, str]:
    """Extract the name information from an ORCID selection

    Parameters
    ----------
    user_orcid : str
        ORCID to search

    Returns
    -------
    user_info : typing.Dict
        dictionary of extracted name metadata
    str
        default output namespace
    """
    _user_info = fdp_id.check_orcid(user_orcid.strip())

    while not _user_info:
        click.echo("Invalid ORCID given.")
        user_orcid = click.prompt("ORCID")
        _user_info = fdp_id.check_orcid(user_orcid.strip())
    
    _user_info['orcid'] = user_orcid

    click.echo(
        f"Found entry: {_user_info['given_names']} "
        f"{_user_info['family_name']}"
    )

    _def_ospace = _user_info['given_names'][0]

    if len(_user_info['family_name'].split()) > 1:
        _def_ospace += _user_info['family_name'].split()[-1]
    else:
        _def_ospace += _user_info['family_name']

    return _user_info, _def_ospace


def _handle_ror(user_ror: str) -> typing.Tuple[typing.Dict, str]:
    """Extract institution name information from an ROR ID

    Parameters
    ----------
    user_ror : str
        ROR ID for an institution

    Returns
    -------
    user_info : typing.Dict
        dictionary of extracted name metadata
    str
        default output namespace
    """
    _user_info = fdp_id.check_ror(user_ror.strip())

    while not _user_info:
        click.echo("Invalid ROR ID given.")
        user_ror = click.prompt("ROR ID")
        _user_info = fdp_id.check_ror(user_ror.strip())

    _user_info['ror'] = user_ror

    click.echo(
        f"Found entry: {_user_info['family_name']} "
    )

    _def_ospace = _user_info['family_name'].lower().replace(' ', '_')

    return _user_info, _def_ospace


def _handle_grid(user_grid: str) -> typing.Tuple[typing.Dict, str]:
    """Extract institution name information from an GRID ID

    Parameters
    ----------
    user_grid : str
        GRID ID for an institution

    Returns
    -------
    user_info : typing.Dict
        dictionary of extracted name metadata
    str
        default output namespace
    """
    _user_info = fdp_id.check_grid(user_grid.strip())

    while not _user_info:
        click.echo("Invalid GRID ID given.")
        user_grid = click.prompt("GRID ID")
        _user_info = fdp_id.check_grid(user_grid.strip())

    _user_info['grid'] = user_grid

    click.echo(
        f"Found entry: {_user_info['family_name']} "
    )

    _def_ospace = _user_info['family_name'].lower().replace(' ', '_')

    return _user_info, _def_ospace


def _handle_uuid() -> typing.Tuple[typing.Dict, str]:
    """Obtain metadata for user where no ID provided

    Returns
    -------
    user_info : typing.Dict
        dictionary of extracted name metadata
    str
        default output namespace
    """
    _full_name = click.prompt("Full Name")
    _def_ospace = ""
    _user_info = {}
    if len(_full_name.split()) > 1:
        _given_name, _family_name = _full_name.split(" ", 1)
        _def_ospace = _full_name.lower().strip()[0]
        _def_ospace += _full_name.lower().split()[-1]
        _user_info['given_names'] = _given_name.strip()
        _user_info['family_name'] = _family_name.strip()
    else:
        _def_ospace += _full_name
        _user_info['given_names'] = _full_name
        _user_info['family_name'] = None

    return _user_info, _def_ospace


def _get_user_info_and_namespaces() -> typing.Dict[str, typing.Dict]:
    _user_email = click.prompt("Email")

    _invalid_input = True

    while _invalid_input:
        _id_type = click.prompt("Use ID (ORCID/ROR/GRID)", default="None")

        if _id_type.upper() == "ORCID":
            _user_orcid = click.prompt("ORCID")
            _user_info, _def_ospace = _handle_orcid(_user_orcid)
            _invalid_input = False
        elif _id_type.upper() == "ROR":
            _user_ror = click.prompt("ROR ID")
            _user_info, _def_ospace = _handle_ror(_user_ror)
            _invalid_input = False
        elif _id_type.upper() == "GRID":
            _user_grid = click.prompt("GRID ID")
            _user_info, _def_ospace = _handle_grid(_user_grid)
            _invalid_input = False
        elif _id_type.upper() == "NONE":
            _user_info, _def_ospace = _handle_uuid()
            _user_uuid = str(uuid.uuid4())
            _user_info['uuid'] = _user_uuid
            _invalid_input = False

    _user_info['email'] = _user_email

    _def_ospace = _def_ospace.lower().replace(" ", "").strip()

    _def_ispace = click.prompt("Default input namespace", default="None")
    _def_ispace = _def_ispace if _def_ispace != "None" else None
    _def_ospace = click.prompt(
        "Default output namespace", default=_def_ospace
    )

    _namespaces = {"input": _def_ispace, "output": _def_ospace}

    return {"user": _user_info, "namespaces": _namespaces}


def global_config_query(registry: str = None) -> typing.Dict[str, typing.Any]:
    """Ask user question set for creating global FAIR config"""
    if not registry:
        registry = fdp_serv.DEFAULT_REGISTRY_LOCATION
    logger.debug("Running global configuration query with registry at '%s'", registry)
    click.echo("Checking for local registry")
    if check_registry_exists(registry):
        click.echo("Local registry found")
    else:
        _install_reg = click.confirm(
            f"Local registry not found at location '{registry}',"
            "would you like to install now?",
            default=True
        )
        if not _install_reg:
            raise fdp_exc.CLIConfigurationError(
                "FAIR repository initialisation requires a local registry installation, aborting."
            )
        fdp_serv.install_registry(install_dir=registry)

    _local_uri = click.prompt("Local Registry URL", default=fdp_serv.DEFAULT_LOCAL_REGISTRY_URL)

    _default_rem = urljoin(fdp_serv.DEFAULT_REGISTRY_DOMAIN, 'api/')
    _remote_url = click.prompt("Remote API URL", default=_default_rem)

    _rem_data_store = click.prompt(
        "Remote Data Storage Root",
        default=_remote_url.replace("api", "data")
    )

    _default_token = os.path.join(registry, "remote-token")
    _rem_key_file = click.prompt(
        "Remote API Token File",
        default=_default_token
    )
    _rem_key_file = os.path.expandvars(_rem_key_file)

    #TODO fix search for valid token
    while (
        False and
        (not os.path.exists(_rem_key_file)
        or not open(_rem_key_file).read().strip()
        )):
        click.echo(
            f"Token file '{_rem_key_file}' does not exist or is empty, "
            "please provide a valid token file."
        )
        _rem_key_file = click.prompt("Remote API Token File")
        _rem_key_file = os.path.expandvars(_rem_key_file)

    if not fdp_serv.check_server_running(_local_uri):
        _run_server = click.confirm(
            "Local registry is offline, would you like to start it?",
            default=False
        )
        if _run_server:
            fdp_serv.launch_server(_local_uri, registry_dir=registry)

            # Keep server running by creating user run cache file
            _cache_addr = os.path.join(fdp_com.session_cache_dir(), 'user.run')
            pathlib.Path(_cache_addr).touch()

        else:
            click.echo("Temporarily launching server to retrieve API token.")
            fdp_serv.launch_server(_local_uri, registry_dir=registry)
            fdp_serv.stop_server(registry_dir=registry, local_uri=_local_uri)
            try:
                fdp_req.local_token(registry_dir=registry)
            except fdp_exc.FileNotFoundError:
                raise fdp_exc.RegistryError(
                    "Failed to retrieve local API token from registry."
                )

    _loc_data_store = click.prompt(
        "Default Data Store",
        default=os.path.join(fdp_com.USER_FAIR_DIR, 'data/')
    )
    if _loc_data_store[-1] != os.path.sep:
        _loc_data_store += os.path.sep

    _glob_conf_dict = _get_user_info_and_namespaces()
    _glob_conf_dict['registries']  = {
        'local': {
            'uri': _local_uri,
            'directory': os.path.abspath(registry),
            'data_store': _loc_data_store
        },
        'origin': {
            'uri': _remote_url,
            'token': _rem_key_file,
            'data_store': _rem_data_store
        }
    }

    return _glob_conf_dict


def local_config_query(
    global_config: typing.Dict[str, typing.Any] = read_global_fdpconfig(),
    first_time_setup: bool = False,
) -> typing.Dict[str, typing.Any]:
    """Ask user questions to create local user config

    Parameters
    ----------
    global_config : Dict[str, Any], optional
        global configuration dictionary
    first_time_setup : bool, optional
        if first time need to setup globals as well, by default False

    Returns
    -------
    Dict[str, Any]
        dictionary of local configurations
    """
    # Try extracting global configurations. If any keys do not exist re-run
    # setup for creation of these, then try again.
    try:
        _def_remote = global_config['registries']['origin']['uri']
        _def_rem_key = global_config['registries']['origin']['token']
        _def_ospace = global_config['namespaces']['output']
        _def_user = global_config['user']
    except KeyError:
        click.echo(
            "Error: Failed to read global configuration,"
            " re-running global setup."
        )
        first_time_setup = True
        global_config = global_config_query()
        _def_remote = global_config['registries']['origin']['uri']
        _def_rem_key = global_config['registries']['origin']['token']
        _def_ospace = global_config['namespaces']['output']
        _def_user = global_config['user']

    # Allow the user to continue without an input namespace as some
    # functionality does not require this.
    if "input" not in global_config['namespaces'] or not global_config['namespaces']:
        click.echo(
            "Warning: No global input namespace declared,"
            " in order to use the registry you will need to specify one"
            " within this local configuration."
        )
        _def_ispace = None
    else:
        _def_ispace = global_config['namespaces']['input']

    # Try checking to see if the current location is a git repository and
    # suggesting this as a default
    try:
        _git_repo = fdp_com.find_git_root(os.getcwd())
    except fdp_exc.UserConfigError:
        _git_repo = None

    _git_repo = click.prompt("Local Git repository", default=_git_repo)
    _invalid_repo = True

    # Check this is indeed a git repository
    while _invalid_repo:
        try:
            git.Repo(_git_repo)
            _invalid_repo = False
        except (git.InvalidGitRepositoryError, git.NoSuchPathError):
            _invalid_repo = True
            click.echo(
                "Invalid directory, location is not the head of a local"
                " git repository."
            )
            _git_repo = click.prompt("Local Git repository")

    # Set the remote to use within git by label, by default 'origin'
    # check the remote label is valid before proceeding
    try:
        git.Repo(_git_repo).remotes['origin']
        _def_rem = 'origin'
    except IndexError:
        _def_rem = None

    _git_remote = click.prompt("Git remote name", default=_def_rem)
    _invalid_rem = True

    while _invalid_rem and _def_rem:
        try:
            git.Repo(_git_repo).remotes[_git_remote]
            _invalid_rem = False
        except IndexError:
            _invalid_rem = True
            click.echo("Invalid remote name for git repository")
            _git_remote = click.prompt("Git remote name", default=_def_rem)

    _repo = git.Repo(_git_repo)

    while _git_remote not in _repo.remotes:
        click.echo(
            f"Git remote label '{_git_remote}' does not exist"
        )
        _git_remote = click.prompt("Git remote name")

    _git_remote_repo = _repo.remotes[_git_remote].url

    click.echo(
        f"Using git repository remote '{_git_remote}': "
        f"{_git_remote_repo}"
    )

    # If this is not the first setup it means globals are available so these
    # can be suggested as defaults during local setup
    if not first_time_setup:
        _def_remote = click.prompt("Remote API URL", default=_def_remote)
        _def_rem_key = click.prompt("Remote API Token File", default=_def_rem_key)
        _def_rem_key = os.path.expandvars(_def_rem_key)

        #TODO fix search for valid token
        while (
            False
            #not os.path.exists(_def_rem_key)
            #or not open(_def_rem_key).read().strip()
        ):
            click.echo(
                f"Token file '{_def_rem_key}' does not exist or is empty, "
                "please provide a valid token file."
            )
            _def_rem_key = click.prompt("Remote API Token File")
            _def_rem_key = os.path.expandvars(_def_rem_key)
        _def_ospace = click.prompt(
            "Default output namespace", default=_def_ospace
        )
        _def_ispace = click.prompt(
            "Default input namespace", default=_def_ispace
        )

    _local_config: typing.Dict[str, typing.Any] = {}

    _local_config['namespaces'] = {
        "output": _def_ospace,
        "input": _def_ispace,
    }

    _local_config['git'] = {
        "remote": _git_remote,
        "local_repo": _git_repo,
        "remote_repo": _git_remote_repo
    }

    # Copy the global configuration then substitute updated
    # configurations
    _local_config['registries'] = copy.deepcopy(global_config['registries'])

    # Local registry is a globally defined entity
    del _local_config['registries']['local']

    _local_config['registries']['origin']['uri'] =  _def_remote
    
    _local_config['user'] = _def_user

    return _local_config

